from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.report_base import ReportBaseAttributes
from pytpp.attributes.report_filter_base import ReportFilterBaseAttributes


class ReportExpirationAttributes(ReportBaseAttributes, ReportFilterBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Report:Expiration"
	options = Attribute('Options')



from typing import List
from pytpp.api.api_base import API, APIResponse, api_response_property
from pytpp.properties.response_objects.ssh_certificates import SSHCertificate
from pytpp.tools.helpers.date_converter import from_date_string


class _SSHCertificates:
    def __init__(self, api_obj):
        self.CAKeyPair = self._CAKeyPair(api_obj=api_obj)
        self.Request = self._Request(api_obj=api_obj)
        self.Retrieve = self._Retrieve(api_obj=api_obj)
        self.Template = self._Template(api_obj=api_obj)

    class _CAKeyPair:
        def __init__(self, api_obj):
            self.Create = self._Create(api_obj=api_obj)
        
        class _Create(API):
            def __init__(self, api_obj):
                super().__init__(api_obj=api_obj, url='SSHCertificates/CAKeyPair/Create')
            
            def post(self, name: str, parent_dn: str = None, key_algorithm: str = None,
                     key_storage: str = None, private_key_data: str = None,
                     private_key_passphrase: str = None):
                body = {
                    'Name': name,
                    'ParentDN': parent_dn,
                    'KeyAlgorithm': key_algorithm,
                    'KeyStorage': key_storage,
                    'PrivateKeyData': private_key_data,
                    'PrivateKeyPassphrase': private_key_passphrase
                }
                
                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def created_on(self):
                        return from_date_string(self._from_json(key='CreatedOn'))

                    @property
                    @api_response_property()
                    def dn(self) -> str:
                        return self._from_json(key='DN')

                    @property
                    @api_response_property()
                    def fingerprint_sha_256(self) -> str:
                        return self._from_json(key='FingerprintSHA256')

                    @property
                    @api_response_property()
                    def guid(self) -> str:
                        return self._from_json(key='Guid')

                    @property
                    @api_response_property()
                    def key_algorithm(self) -> str:
                        return self._from_json(key='KeyAlgorithm')

                    @property
                    @api_response_property()
                    def key_storage(self) -> str:
                        return self._from_json(key='KeyStorage')

                    @property
                    @api_response_property()
                    def name(self) -> str:
                        return self._from_json(key='Name')

                    @property
                    @api_response_property()
                    def processing_details(self):
                        return SSHCertificate.ProcessingDetails(self._from_json(key='ProcessingDetails'))

                    @property
                    @api_response_property()
                    def public_key_data(self) -> str:
                        return self._from_json(key='PublicKeyData')

                    @property
                    @api_response_property()
                    def response(self):
                        return SSHCertificate.Response(self._from_json(key='Response'))
                        
                return _Response(response=self._post(data=body))

    class _Request(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSHCertificates/Request')

        def post(self, ca_dn: str,
                 key_id: str,
                 destination_address: str = None,
                 extensions: str = None,
                 force_command: str = None,
                 object_name: str = None,
                 origin: str = None,
                 policy_dn: str = None,
                 principals: List[str] = None,
                 public_key_data: str = None,
                 source_addresses: List[str] = None,
                 validity_period: str = None
                 ):

            body = {
                'CADN': ca_dn,
                'KeyId': key_id,
                'DestinationAddress': destination_address,
                'Extensions': extensions,
                'ForceCommand': force_command,
                'ObjectName': object_name,
                'Origin': origin,
                'PolicyDN': policy_dn,
                'Principals': principals,
                'PublicKeyData': public_key_data,
                'SourceAddresses': source_addresses,
                'ValidityPeriod': validity_period
            }
            

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def ssh_certificate_dn(self) -> str:
                    return self._from_json('DN')

                @property
                @api_response_property()
                def ssh_certificate_guid(self) -> str:
                    return self._from_json('Guid')

                @property
                @api_response_property()
                def response(self) -> SSHCertificate.Response:
                    return SSHCertificate.Response(self._from_json('Response'))

                @property
                @api_response_property()
                def processing_details(self) -> SSHCertificate.ProcessingDetails:
                    return SSHCertificate.ProcessingDetails(self._from_json('ProcessingDetails'))


            return _Response(response=self._post(data=body))

    class _Retrieve(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSHCertificates/Retrieve')

        def post(self, certificate_dn: str = None,
                 certificate_guid: str = None,
                 include_certificate_details: bool = None,
                 include_private_key_data: bool = None,
                 private_key_passphrase: str = None):

            body = {
                'DN': certificate_dn,
                'Guid': certificate_guid,
                'IncludeCertificateDetails': include_certificate_details,
                'IncludePrivateKeyData': include_private_key_data,
                'PrivateKeyPassphrase': private_key_passphrase
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def ca_dn(self) -> str:
                    return self._from_json('CADN')

                @property
                @api_response_property()
                def ca_guid(self) -> str:
                    return self._from_json('CAGuid')

                @property
                @api_response_property()
                def certificate_data(self) -> str:
                    return self._from_json('CertificateData')

                @property
                @api_response_property()
                def certificate_details(self) -> SSHCertificate.CertificateDetails:
                    return SSHCertificate.CertificateDetails(self._from_json('CertificateDetails'))

                @property
                @api_response_property()
                def guid(self) -> str:
                    return self._from_json('Guid')

                @property
                @api_response_property()
                def key_id(self) -> str:
                    return self._from_json('KeyID')

                @property
                @api_response_property()
                def private_key_data(self) -> str:
                    return self._from_json('PrivateKeyData')

                @property
                @api_response_property()
                def processing_details(self) -> SSHCertificate.ProcessingDetails:
                    return SSHCertificate.ProcessingDetails(self._from_json('ProcessingDetails'))

                @property
                @api_response_property()
                def public_key_data(self) -> str:
                    return self._from_json('PublicKeyData')

                @property
                @api_response_property()
                def request_details(self) -> SSHCertificate.RequestDetails:
                    return SSHCertificate.RequestDetails(self._from_json('RequestDetails'))

                @property
                @api_response_property()
                def response(self):
                    return SSHCertificate.Response(self._from_json('Response'))

            return _Response(response=self._post(data=body))

    class _Template(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SSHCertificates/Template')
            self.Retrieve = self._Retrieve(api_obj=api_obj)

        class _Retrieve(API):
            def __init__(self, api_obj):
                super().__init__(api_obj=api_obj, url='/SSHCertificates/Template/Retrieve')
                self.PublicKeyData = self._PublicKeyData(api_obj=api_obj)

            def post(self, template_dn: str = None, template_guid: str = None, include_ca_keypair_details: bool = None):
                body = {
                    'DN': template_dn,
                    'Guid'  : template_guid,
                    'IncludeCAKeyPairDetails' : include_ca_keypair_details
                }

                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def access_control(self) -> dict:
                        return self._from_json('AccessControl')

                    @property
                    @api_response_property()
                    def api_client(self) -> SSHCertificate.APIClient:
                        return SSHCertificate.APIClient(self._from_json('APIClient'))

                    @property
                    @api_response_property()
                    def ca_keypair_guid(self) -> str:
                        return self._from_json('CAKeyPairGuid')

                    @property
                    @api_response_property()
                    def ca_keypair_dn(self) -> str:
                        return self._from_json('CAKeyPairDN')

                    @property
                    @api_response_property()
                    def ca_keypair(self) -> SSHCertificate.CAKeyPair:
                        return SSHCertificate.CAKeyPair(self._from_json('CAKeyPair'))

                    @property
                    @api_response_property()
                    def certificate(self) -> SSHCertificate.Certificate:
                        return SSHCertificate.Certificate(self._from_json('Certificate'))

                    @property
                    @api_response_property()
                    def contacts(self) -> List[str]:
                        return self._from_json('Contacts')

                    @property
                    @api_response_property()
                    def created_on(self) -> str:
                        return self._from_json('CreatedOn')

                    @property
                    @api_response_property()
                    def guid(self) -> str:
                        return self._from_json('Guid')

                    @property
                    @api_response_property()
                    def name(self) -> str:
                        return self._from_json('Name')

                    @property
                    @api_response_property()
                    def response(self):
                        return SSHCertificate.Response(self._from_json('Response'))

                return _Response(response=self._post(data=body))

            class _PublicKeyData(API):
                def __init__(self, api_obj):
                    super().__init__(api_obj=api_obj, url='SSHCertificates/Template/Retrieve/PublicKeyData')

                def get(self, template_dn: str = None, template_guid: str = None):
                    params = {
                        'DN'  : template_dn,
                        'Guid': template_guid,
                    }

                    class _Response(APIResponse):
                        def __init__(self, response):
                            super().__init__(response=response)

                        @property
                        def response(self) -> str:
                            return self.response

                    return _Response(response=self._get(params=params))



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class LogServerAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Log Server"
	log_application_container = Attribute('Log Application Container')
	log_channel_container = Attribute('Log Channel Container')
	log_notification_container = Attribute('Log Notification Container')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.client_work_base import ClientWorkBaseAttributes


class ClientAgentConfigurationWorkAttributes(ClientWorkBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Client Agent Configuration Work"
	agent_cert_trust_bundle = Attribute('Agent Cert Trust Bundle')
	days_of_month = Attribute('Days Of Month')
	days_of_week = Attribute('Days Of Week')
	event_row_count = Attribute('Event Row Count', min_version='15.4')
	interval = Attribute('Interval')
	log_communications = Attribute('Log Communications', min_version='15.4')
	log_facility = Attribute('Log Facility')
	log_threshold = Attribute('Log Threshold')
	minimum_free_space = Attribute('Minimum Free Space')
	proxy_credential = Attribute('Proxy Credential')
	proxy_host = Attribute('Proxy Host')
	schedule_type = Attribute('Schedule Type')
	start_time = Attribute('Start Time')
	tls_crl_verify = Attribute('TLS CRL Verify', min_version='15.4')
	web_service_url = Attribute('Web Service URL')



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.top import TopAttributes


class ClientWorkRootAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Client Work Root"



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class F5Attributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "F5"
	bundle_certificate = Attribute('Bundle Certificate')
	certificate_chain_name = Attribute('Certificate Chain Name')
	certificate_name = Attribute('Certificate Name')
	file_validation_disabled = Attribute('File Validation Disabled')
	fips_key = Attribute('Fips Key')
	network_validation_disabled = Attribute('Network Validation Disabled')
	overwrite_certificate = Attribute('Overwrite Certificate')
	overwrite_existing_chain = Attribute('Overwrite Existing Chain')
	overwrite_key = Attribute('Overwrite Key')
	private_key_name = Attribute('Private Key Name')
	private_key_type = Attribute('Private Key Type')
	ssh_port = Attribute('SSH Port')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class WorkflowAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Workflow"
	rule = Attribute('Rule')
	rule_vault_id = Attribute('Rule Vault Id', min_version='19.3')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.flow_action_base import FlowActionBaseAttributes


class AdaptableFlowActionBaseAttributes(FlowActionBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Adaptable Flow Action Base"
	adaptable_script_max_file_size = Attribute('Adaptable Script Max File Size', min_version='21.4')
	adaptable_script_try_later_max_attempts = Attribute('Adaptable Script Try Later Max Attempts', min_version='21.4')
	adaptable_scripts_directory_name = Attribute('Adaptable Scripts Directory Name', min_version='21.4')
	log_debug = Attribute('Log Debug', min_version='21.4')
	oauth_token_application_id = Attribute('OAuth Token Application Id', min_version='21.4')
	oauth_token_credential = Attribute('OAuth Token Credential', min_version='21.4')
	oauth_token_scope = Attribute('OAuth Token Scope', min_version='21.4')
	powershell_script = Attribute('PowerShell Script', min_version='21.4')
	powershell_script_hash_vault_id = Attribute('PowerShell Script Hash Vault Id', min_version='21.4')
	script_execution_timeout = Attribute('Script Execution Timeout', min_version='21.4')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.driver_base import DriverBaseAttributes


class AgentDriverBaseAttributes(DriverBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Agent Driver Base"
	address = Attribute('Address')
	agent_discovery_redirection_rule = Attribute('Agent Discovery Redirection Rule')
	credential = Attribute('Credential')
	expiration = Attribute('Expiration')
	extended_host_details = Attribute('Extended Host Details')
	interval = Attribute('Interval')
	log_threshold = Attribute('Log Threshold')
	managed_by_discovery_dn = Attribute('Managed By Discovery DN')
	module_disabled = Attribute('Module Disabled')
	port = Attribute('Port')
	version = Attribute('Version')



from pytpp.attributes._helper import IterableMeta, Attribute


class DiscoveryStatisticsAttributes(metaclass=IterableMeta):
	__config_class__ = "Discovery Statistics"
	certificates_found = Attribute('Certificates Found')
	completed_scans = Attribute('Completed Scans')
	connect_succeeded = Attribute('Connect Succeeded')
	keys_found = Attribute('Keys Found')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.certificate_trust_store_base import CertificateTrustStoreBaseAttributes


class F5LTMAdvancedTrustStoreAttributes(CertificateTrustStoreBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "F5 LTM Advanced Trust Store"
	certificate_name = Attribute('Certificate Name')
	ssh_port = Attribute('SSH Port')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.flow_action_approver_base import FlowActionApproverBaseAttributes


class FlowActionConfigReadApproversAttributes(FlowActionApproverBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Flow Action Config Read Approvers"
	attribute_name = Attribute('Attribute Name', min_version='19.2')
	object_dn = Attribute('Object DN', min_version='19.2')
	policy_read = Attribute('Policy Read', min_version='19.2')
	use_administrators_as_fallback_approvers = Attribute('Use Administrators as Fallback Approvers', min_version='20.2')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.schedule_base import ScheduleBaseAttributes
from pytpp.attributes.top import TopAttributes


class CAImportAttributes(ScheduleBaseAttributes, TopAttributes, metaclass=IterableMeta):
	__config_class__ = "CA Import"
	amazon_private_ca = Attribute('Amazon Private CA', min_version='21.2')
	assign_contact_value_to_issued_to = Attribute('Assign Contact value to Issued To', min_version='19.1')
	automatically_import = Attribute('Automatically Import')
	aws_credential_dn = Attribute('Aws Credential DN', min_version='21.2')
	bypass_reconciliation = Attribute('Bypass Reconciliation', min_version='15.1')
	ca_type = Attribute('CA Type')
	certificate_credential = Attribute('Certificate Credential', min_version='19.1')
	certificate_origin = Attribute('Certificate Origin', min_version='19.2')
	credential = Attribute('Credential')
	default_container = Attribute('Default Container')
	discover_all = Attribute('Discover All')
	driver_name = Attribute('Driver Name')
	f5_ltm_advanced_hostname = Attribute('F5 LTM Advanced:Hostname', min_version='15.2')
	import_summary = Attribute('Import Summary', min_version='19.1')
	in_progress = Attribute('In Progress')
	include_expired = Attribute('Include Expired')
	include_revoked = Attribute('Include Revoked')
	last_processed_id = Attribute('Last Processed Id', min_version='19.1')
	last_run = Attribute('Last Run')
	last_update = Attribute('Last Update')
	log_debug = Attribute('Log Debug', min_version='19.3')
	microsoft_ca_hostname = Attribute('Microsoft CA:Hostname')
	microsoft_ca_query_by_template_name = Attribute('Microsoft CA:Query By Template Name', min_version='15.3')
	microsoft_ca_request_id = Attribute('Microsoft CA:Request ID')
	microsoft_ca_request_id_by_template_name = Attribute('Microsoft CA:Request ID By Template Name', min_version='15.3')
	microsoft_ca_service_name = Attribute('Microsoft CA:Service Name')
	microsoft_ca_template = Attribute('Microsoft CA:Template')
	page_size = Attribute('Page Size', min_version='16.4')
	placement_rule = Attribute('Placement Rule')
	placement_rule_item = Attribute('Placement Rule Item')
	placement_summary = Attribute('Placement Summary')
	powershell_script = Attribute('PowerShell Script', min_version='19.1')
	powershell_script_hash_vault_id = Attribute('PowerShell Script Hash Vault Id', min_version='19.1')
	profile_string = Attribute('Profile String', min_version='20.1')
	region_code = Attribute('Region Code', min_version='21.2')
	result_count = Attribute('Result Count')
	s3_bucket = Attribute('S3 Bucket', min_version='21.2')
	script_execution_timeout = Attribute('Script Execution Timeout', min_version='20.2')
	script_hash_mismatch_error = Attribute('Script Hash Mismatch Error', min_version='19.1')
	secondary_credential = Attribute('Secondary Credential', min_version='19.1')
	service_address = Attribute('Service Address', min_version='20.1')
	status = Attribute('Status')
	stop_requested = Attribute('Stop Requested')
	template_name_certificate_origin = Attribute('Template Name Certificate Origin', min_version='19.2')



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.agent_discovery_base import AgentDiscoveryBaseAttributes


class AgentSSHDiscoveryAttributes(AgentDiscoveryBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Agent SSH Discovery"



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.flow_action_base import FlowActionBaseAttributes


class CodeSignPreApprovalActionAttributes(FlowActionBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Code Sign PreApproval Action"
	stage_identifier = Attribute('Stage Identifier', min_version='20.2')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.code_signing_environment_base import CodeSigningEnvironmentBaseAttributes


class CodeSigningGPGEnvironmentAttributes(CodeSigningEnvironmentBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Code Signing GPG Environment"
	authentication_key_algorithm = Attribute('Authentication Key Algorithm', min_version='20.3')
	authentication_key_dn = Attribute('Authentication Key DN', min_version='20.2')
	encryption_key_algorithm = Attribute('Encryption Key Algorithm', min_version='20.3')
	encryption_key_dn = Attribute('Encryption Key DN', min_version='20.2')
	internet_email_address = Attribute('Internet EMail Address', min_version='20.3')
	issuer_key_dn = Attribute('Issuer Key DN', min_version='21.4')
	key_storage_location = Attribute('Key Storage Location', min_version='20.3')
	max_uses = Attribute('Max Uses', min_version='20.3')
	real_name = Attribute('Real Name', min_version='20.3')
	signing_key_algorithm = Attribute('Signing Key Algorithm', min_version='20.3')
	signing_key_dn = Attribute('Signing Key DN', min_version='20.2')
	validity_period = Attribute('Validity Period', min_version='20.3')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.credential_base import CredentialBaseAttributes


class AdaptableCredentialAttributes(CredentialBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Adaptable Credential"
	credential_connector = Attribute('Credential Connector', min_version='21.1')
	credential_type = Attribute('Credential Type', min_version='21.1')
	option_1 = Attribute('Option 1', min_version='21.1')
	option_2 = Attribute('Option 2', min_version='21.1')
	password_1 = Attribute('Password 1', min_version='21.1')
	text_field_1 = Attribute('Text Field 1', min_version='21.1')
	text_field_2 = Attribute('Text Field 2', min_version='21.1')
	text_field_3 = Attribute('Text Field 3', min_version='21.1')
	text_field_4 = Attribute('Text Field 4', min_version='21.1')
	text_field_5 = Attribute('Text Field 5', min_version='21.1')



from pytpp.properties.config import (
    ApplicationAttributeValues, ClientGroupsAttributeValues, CertificateAttributeValues, CertificateAuthorityAttributeValues,
    DeviceAttributeValues, DiscoveryAttributeValues, CustomFieldAttributeValues, IdentityAttributeValues, WorkflowAttributeValues,
    PlacementRulesAttributeValues, ClientWorkAttributeValues, JumpServerAttributeValues
)

class AttributeValues:
    Application = ApplicationAttributeValues
    Certificate = CertificateAttributeValues
    CertificateAuthority = CertificateAuthorityAttributeValues
    ClientGroups = ClientGroupsAttributeValues
    ClientWork = ClientWorkAttributeValues
    CustomField = CustomFieldAttributeValues
    Device = DeviceAttributeValues
    Discovery = DiscoveryAttributeValues
    Identity = IdentityAttributeValues
    JumpServer = JumpServerAttributeValues
    PlacementRules = PlacementRulesAttributeValues
    Workflow = WorkflowAttributeValues



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.client_agent_ssh_provisioning_work import ClientAgentSSHProvisioningWorkAttributes


class ClientAgentSSHKeyUsageWorkAttributes(ClientAgentSSHProvisioningWorkAttributes, metaclass=IterableMeta):
	__config_class__ = "Client Agent SSH Key Usage Work"
	max_row_count = Attribute('Max Row Count')



from pytpp.attributes._helper import IterableMeta, Attribute


class AgentSSHBaseAttributes(metaclass=IterableMeta):
	__config_class__ = "Agent SSH Base"
	credential = Attribute('Credential')
	exclude_remote_mount_points = Attribute('Exclude Remote Mount Points')
	expiration = Attribute('Expiration')
	file_scan_disabled = Attribute('File Scan Disabled')
	log_threshold = Attribute('Log Threshold')
	max_filesize = Attribute('Max Filesize')
	min_filesize = Attribute('Min Filesize')
	port_validation_disabled = Attribute('Port Validation Disabled')
	ssh_scanner_service_path = Attribute('SSH Scanner Service Path')
	ssh_scanner_user_path = Attribute('SSH Scanner User Path')
	sshd_filename_filter = Attribute('SSHd Filename Filter')
	sshd_max_filesize = Attribute('SSHd Max Filesize')
	server_path_defaults_disabled = Attribute('Server Path Defaults Disabled')
	upload_private_keys_disabled = Attribute('Upload Private Keys Disabled')
	user_path_defaults_disabled = Attribute('User Path Defaults Disabled')



from dataclasses import dataclass


@dataclass
class Identity:
    full_name: str
    is_container: bool
    is_group: bool
    name: str
    prefix: str
    prefixed_name: str
    prefixed_universal: str
    type: str
    universal: str


@dataclass
class InvalidIdentity:
    prefix: str
    prefixed_name: str
    prefixed_universal: str
    universal: str



from dataclasses import dataclass
from typing import List


@dataclass
class Result:
    code: int
    metadata_result: str


@dataclass
class Item:
    allowed_characters: str
    allowed_values: str
    category: str
    classes: list
    config_attribute: str
    date_only: bool
    default_values: str
    display_after: str
    dn: str
    error_message: str
    guid: str
    help: str
    label: str
    localization_table: str
    localized_help: str
    localized_label: str
    localized_set: str
    mandatory: bool
    name: str
    mask: str
    maximum_length: int
    minimum_length: int
    policyable: bool
    regular_expression: str
    render_hidden: bool
    render_read_only: bool
    single: bool
    time_only: bool
    type: int


@dataclass
class Data:
    key: 'Item'
    value: list


@dataclass
class PolicyItem:
    key: str
    value: 'List[Item]'



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.organization import OrganizationAttributes


class IdentityRootAttributes(OrganizationAttributes, metaclass=IterableMeta):
	__config_class__ = "Identity Root"
	authentication_scheme = Attribute('Authentication Scheme')
	identity_cache_timeout = Attribute('Identity Cache Timeout', min_version='16.2')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.driver_base import DriverBaseAttributes


class CertificateAuthorityBaseAttributes(DriverBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Certificate Authority Base"
	additional_field = Attribute('Additional Field')
	concurrent_connection_limit = Attribute('Concurrent Connection Limit')
	credential = Attribute('Credential')
	credits = Attribute('Credits')
	credits_alert = Attribute('Credits Alert')
	credits_used = Attribute('Credits Used')
	enhanced_key_usage = Attribute('Enhanced Key Usage', min_version='19.2')
	host = Attribute('Host')
	manual_approval = Attribute('Manual Approval')
	port = Attribute('Port')
	protection_key = Attribute('Protection Key')
	renewal_window = Attribute('Renewal Window')
	retry_count = Attribute('Retry Count')
	retry_interval = Attribute('Retry Interval')
	san_enabled = Attribute('SAN Enabled')
	signature_algorithm = Attribute('Signature Algorithm')
	specific_end_date_enabled = Attribute('Specific End Date Enabled')
	template = Attribute('Template')
	test_account = Attribute('Test Account')
	timeout = Attribute('Timeout')
	validity_period = Attribute('Validity Period')
	vault_id = Attribute('Vault Id')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.proxy_base import ProxyBaseAttributes
from pytpp.attributes.top import TopAttributes
from pytpp.attributes.zone_base import ZoneBaseAttributes


class VenafiPlatformAttributes(ProxyBaseAttributes, TopAttributes, ZoneBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Venafi Platform"
	acme_fqdn = Attribute('ACME FQDN', min_version='17.2')
	aws_ec2_role_authorized_identities = Attribute('AWS EC2 Role Authorized Identities', min_version='20.4')
	activation_latency_threshold = Attribute('Activation Latency Threshold', min_version='21.3')
	activation_query_frequency = Attribute('Activation Query Frequency', min_version='21.3')
	activation_work_threshold = Attribute('Activation Work Threshold', min_version='21.3')
	allowed_protocol_version = Attribute('Allowed Protocol Version', min_version='16.2')
	aperture_fqdn = Attribute('Aperture FQDN', min_version='16.2')
	authserver_fqdn = Attribute('AuthServer FQDN', min_version='19.1')
	capabilities_asymmetric_key_generation = Attribute('Capabilities: Asymmetric Key Generation')
	certificate_origin = Attribute('Certificate Origin', min_version='19.1')
	certificate_verification = Attribute('Certificate Verification', min_version='17.2')
	certificate_verification_log_warnings = Attribute('Certificate Verification Log Warnings', min_version='18.1')
	check_crl = Attribute('Check CRL', min_version='17.2')
	client_enrollment_require_windows_authentication = Attribute('Client Enrollment Require Windows Authentication', min_version='15.2')
	client_id = Attribute('Client ID', min_version='20.3')
	current_upgrade_task = Attribute('Current Upgrade Task', min_version='20.1')
	disable_triggers = Attribute('Disable Triggers', min_version='17.2')
	display_name = Attribute('Display Name', min_version='20.1')
	est_options = Attribute('Est Options', min_version='20.1')
	hsm_fqdn = Attribute('HSM FQDN', min_version='19.2')
	heartbeat_interval = Attribute('Heartbeat Interval')
	ignore_identity = Attribute('Ignore Identity', min_version='21.2')
	interval = Attribute('Interval')
	keyserver_fqdn = Attribute('KeyServer FQDN', min_version='21.4')
	keystore_cache_check = Attribute('KeyStore Cache Check', min_version='20.4')
	keystore_cache_lifetime = Attribute('KeyStore Cache Lifetime', min_version='20.4')
	keystore_cache_size = Attribute('KeyStore Cache Size', min_version='20.4')
	log_debug = Attribute('Log Debug')
	log_server = Attribute('Log Server')
	log_target = Attribute('Log Target')
	maximum_above_normal_priority_threads = Attribute('Maximum Above Normal Priority Threads', min_version='18.4')
	maximum_below_normal_priority_threads = Attribute('Maximum Below Normal Priority Threads', min_version='18.4')
	maximum_highest_priority_threads = Attribute('Maximum Highest Priority Threads', min_version='18.4')
	maximum_low_priority_threads = Attribute('Maximum Low Priority Threads', min_version='17.4')
	maximum_normal_priority_threads = Attribute('Maximum Normal Priority Threads', min_version='18.4')
	maximum_threads = Attribute('Maximum Threads')
	migration_task = Attribute('Migration Task', min_version='19.2')
	minimum_threads = Attribute('Minimum Threads')
	operational_certificate_dn = Attribute('Operational Certificate DN')
	options = Attribute('Options')
	pending_migration_task = Attribute('Pending Migration Task', min_version='19.2')
	portal_fqdn = Attribute('Portal FQDN', min_version='16.2')
	private_key_vault_id = Attribute('Private Key Vault Id', min_version='20.1')
	processing_enabled = Attribute('Processing Enabled', min_version='21.3')
	scep_fqdn = Attribute('SCEP FQDN', min_version='16.2')
	scep_allowed_identities = Attribute('Scep Allowed Identities')
	scep_allowed_identities_for_http_authentication = Attribute('Scep Allowed Identities For Http Authentication', min_version='21.3')
	scep_certificate_authority = Attribute('Scep Certificate Authority')
	scep_challenge_password = Attribute('Scep Challenge Password')
	scep_encryption_certificate_authority = Attribute('Scep Encryption Certificate Authority')
	scep_encryption_ra_certificate = Attribute('Scep Encryption RA Certificate', min_version='18.3')
	scep_instant_retries_client_user_agent = Attribute('Scep Instant Retries Client User Agent', min_version='19.3')
	scep_intune_application_id = Attribute('Scep Intune Application Id', min_version='19.3')
	scep_intune_application_secret = Attribute('Scep Intune Application Secret', min_version='19.3')
	scep_intune_authentication_authority_resource_url = Attribute('Scep Intune Authentication Authority Resource URL', min_version='19.3')
	scep_intune_challenge_password_sanity_check_disabled = Attribute('Scep Intune Challenge Password Sanity Check Disabled', min_version='19.4')
	scep_intune_graph_resource_url = Attribute('Scep Intune Graph Resource URL', min_version='19.3')
	scep_intune_http_connection_lease_timeout = Attribute('Scep Intune Http Connection Lease Timeout', min_version='19.3')
	scep_intune_http_connection_limit = Attribute('Scep Intune Http Connection Limit', min_version='19.3')
	scep_intune_http_timeout = Attribute('Scep Intune Http Timeout', min_version='19.3')
	scep_intune_provider_name_and_version = Attribute('Scep Intune Provider Name And Version', min_version='19.3')
	scep_intune_resource_url = Attribute('Scep Intune Resource URL', min_version='19.3')
	scep_intune_tenant_name = Attribute('Scep Intune Tenant Name', min_version='19.3')
	scep_intune_validation_attempts = Attribute('Scep Intune Validation Attempts', min_version='19.3')
	scep_maximum_ndes_challenges = Attribute('Scep Maximum Ndes Challenges')
	scep_ndes_challenge_validity = Attribute('Scep Ndes Challenge Validity')
	scep_object_container = Attribute('Scep Object Container')
	scep_options = Attribute('Scep Options')
	scep_ra_certificate = Attribute('Scep RA Certificate')
	scep_signing_certificate_authority = Attribute('Scep Signing Certificate Authority')
	scep_signing_ra_certificate = Attribute('Scep Signing RA Certificate', min_version='18.3')
	standby_mode = Attribute('Standby Mode', min_version='21.3')
	start_time = Attribute('Start Time')
	time_stamping_certificate_dn = Attribute('Time Stamping Certificate DN', min_version='20.3')
	time_stamping_proxy_host_urls = Attribute('Time Stamping Proxy Host URLs')
	time_stamping_proxy_options = Attribute('Time Stamping Proxy Options', min_version='20.3')
	timestampserver_fqdn = Attribute('TimeStampServer FQDN', min_version='20.3')
	upgrade_details = Attribute('Upgrade Details', min_version='20.1')
	upgrade_status = Attribute('Upgrade Status', min_version='20.1')
	user_agent_windows_authentication_enabled = Attribute('User Agent Windows Authentication Enabled', min_version='18.2')
	vedclient_fqdn = Attribute('VEDClient FQDN', min_version='16.2')
	webadmin_fqdn = Attribute('WebAdmin FQDN', min_version='16.2')
	websdk_fqdn = Attribute('WebSDK FQDN', min_version='16.2')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.branch_base import BranchBaseAttributes


class EncryptionRootAttributes(BranchBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Encryption Root"
	encryption_driver = Attribute('Encryption Driver')
	protection_key = Attribute('Protection Key')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.client_work_base import ClientWorkBaseAttributes


class ServerAgentBaseDevicePlacementWorkAttributes(ClientWorkBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Server Agent Base Device Placement Work"
	device_object_location = Attribute('Device Object Location', min_version='20.1')
	device_share_mode = Attribute('Device Share Mode', min_version='20.1')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class CiscoACEAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "CiscoACE"
	certificate_file = Attribute('Certificate File')
	class_map = Attribute('Class Map')
	context_name = Attribute('Context Name')
	file_validation_disabled = Attribute('File Validation Disabled')
	import_only = Attribute('Import Only')
	include_root_chain = Attribute('Include Root Chain')
	network_validation_disabled = Attribute('Network Validation Disabled')
	non_exportable = Attribute('Non-Exportable')
	policy_map = Attribute('Policy Map')
	private_key_file = Attribute('Private Key File')
	sftp_credential = Attribute('SFTP Credential')
	sftp_host = Attribute('SFTP Host')
	sftp_path = Attribute('SFTP Path')
	sftp_port = Attribute('SFTP Port')
	ssl_proxy_service = Attribute('SSL Proxy Service')
	temp_certificate_label = Attribute('Temp Certificate Label')



from base64 import b64encode
from hashlib import sha256
from pytpp.properties.config import ApplicationAttributeValues
from pytpp.attributes.adaptable_app import AdaptableAppAttributes
from pytpp.attributes.amazon_app import AmazonAppAttributes
from pytpp.attributes.apache import ApacheAttributes
from pytpp.attributes.azure_key_vault import AzureKeyVaultAttributes
from pytpp.attributes.bluecoat_sslva import BlueCoatSSLVAAttributes
from pytpp.attributes.capi import CAPIAttributes
from pytpp.attributes.netscaler import NetScalerAttributes
from pytpp.attributes.connectdirect import ConnectDirectAttributes
from pytpp.attributes.f5_authentication_bundle import F5AuthenticationBundleAttributes
from pytpp.attributes.f5_ltm_advanced import F5LTMAdvancedAttributes
from pytpp.attributes.google_cloud_app import GoogleCloudAppAttributes
from pytpp.attributes.datapower import DataPowerAttributes
from pytpp.attributes.gsk import GSKAttributes
from pytpp.attributes.imperva_mx import ImpervaMXAttributes
from pytpp.attributes.jks import JKSAttributes
from pytpp.attributes.iplanet import iPlanetAttributes
from pytpp.attributes.palo_alto_network_fw import PaloAltoNetworkFWAttributes
from pytpp.attributes.pkcs_12 import PKCS12Attributes
from pytpp.attributes.pkcs11 import PKCS11Attributes
from pytpp.attributes.riverbed_steelhead import RiverbedSteelHeadAttributes
from pytpp.attributes.tealeaf_pca import TealeafPCAAttributes
from pytpp.attributes.vam_nshield import VAMnShieldAttributes
from pytpp.attributes.application_base import ApplicationBaseAttributes
from pytpp.attributes.application_group import ApplicationGroupAttributes
from pytpp.attributes.apache_application_group import ApacheApplicationGroupAttributes
from pytpp.attributes.pkcs11_application_group import PKCS11ApplicationGroupAttributes
from pytpp.attributes.x509_certificate import X509CertificateAttributes
from pytpp.features.bases.feature_base import FeatureBase, feature
from pytpp.features.definitions.exceptions import InvalidResultCode, UnexpectedValue, FeatureException
from pytpp.features.definitions.classes import Classes
from pytpp.properties.secret_store import KeyNames, Namespaces, VaultTypes
from pytpp.tools.helpers.date_converter import from_date_string
from typing import Union, List, TYPE_CHECKING
if TYPE_CHECKING:
    from pytpp.tools.vtypes import Config, Identity


# region Applications
class _ApplicationBase(FeatureBase):
    def __init__(self, api, class_name: str):
        super().__init__(api=api)
        self._class_name = str(class_name)

    def delete(self, application: 'Union[Config.Object, str]'):
        """
        Deletes an application object.

        Args:
            application: :ref:`config_object` or :ref:`dn` of the application object.
        """
        application_dn = self._get_dn(application)
        self._secret_store_delete(object_dn=application_dn)
        self._config_delete(object_dn=application_dn)

    def disable(self, application: 'Union[Config.Object, str]'):
        """
        Disables all processing and provisioning of the application.

        Args:
            application: :ref:`config_object` or :ref:`dn` of the application object.
        """
        application_dn = self._get_dn(application)
        result = self._api.websdk.Config.Write.post(
            object_dn=application_dn,
            attribute_data=self._name_value_list({
                ApplicationBaseAttributes.disabled: ["1"]
            }, keep_list_values=True)
        ).result

        if result.code != 1:
            raise InvalidResultCode(code=result.code, code_description=result.config_result)

    def enable(self, application: 'Union[Config.Object, str]'):
        """
        Enables all processing and provisioning of the application.

        Args:
            application: :ref:`config_object` or :ref:`dn` of the application object.
        """
        application_dn = self._get_dn(application)
        result = self._api.websdk.Config.ClearAttribute.post(
            object_dn=application_dn,
            attribute_name=ApplicationBaseAttributes.disabled
        ).result

        if result.code != 1:
            raise InvalidResultCode(code=result.code, code_description=result.config_result)

    def get(self, application_dn: str, raise_error_if_not_exists: bool = True):
        """
        Args:
            application_dn: :ref:`dn` of the application object.
            raise_error_if_not_exists: Raise an exception if the application :ref:`dn` does not exist.

        Returns:
            :ref:`config_object` of the application
        """
        return self._get_config_object(
            object_dn=application_dn,
            raise_error_if_not_exists=raise_error_if_not_exists
        )

    def get_associated_certificate(self, application: 'Union[Config.Object, str]'):
        """
        Args:
            application: :ref:`config_object` or :ref:`dn` of the application object.

        Returns:
            :ref:`config_object` of the certificate object associated to the application object.
        """
        application_dn = self._get_dn(application)
        response = self._api.websdk.Config.Read.post(
            object_dn=application_dn,
            attribute_name=ApplicationBaseAttributes.certificate
        )

        if not response.values:
            return None

        certificate_dn = response.values[0]
        return self._api.websdk.Config.IsValid.post(object_dn=certificate_dn).object

    def _create(self, name: 'str', device: 'Union[Config.Object, str]', description: 'str' = None,
                contacts: 'List[Union[Identity.Identity, str]]' = None,
                approvers: 'List[Union[Identity.Identity, str]]' = None,
                attributes: dict = None, get_if_already_exists: bool = True):
        device_dn = self._get_dn(obj=device)
        if approvers is not None:
            attributes.update({
                ApplicationBaseAttributes.approver: [self._get_prefixed_universal(a) for a in approvers]
            })
        if contacts is not None:
            attributes.update({
                ApplicationBaseAttributes.contact: [self._get_prefixed_universal(c) for c in contacts]
            })
        if description is not None:
            attributes.update({
                ApplicationBaseAttributes.description: description
            })

        return self._config_create(
            name=name,
            parent_folder_dn=device_dn,
            config_class=self._class_name,
            attributes=attributes,
            get_if_already_exists=get_if_already_exists
        )

    def _get_stage(self, application: 'Union[Config.Object, str]'):
        application_dn = self._get_dn(application)
        response = self._api.websdk.Config.Read.post(
            object_dn=application_dn,
            attribute_name=ApplicationBaseAttributes.stage
        )

        return int(response.values[0]) if response.values else None

    def get_stage(self, application: 'Union[Config.Object, str]'):
        """
        Args:
            application: :ref:`config_object` or :ref:`dn` of the application object.

        Returns:
            int: The current stage if it exists. Otherwise, returns ``None``.
        """
        self._get_stage(application=application)

    def get_status(self, application: 'Union[Config.Object, str]'):
        """
        Args:
            application: :ref:`config_object` or :ref:`dn` of the application object.

        Returns:
            str: The current processing status of the application object or ``None`` if a status does not exist.
        """
        application_dn = self._get_dn(application)
        response = self._api.websdk.Config.Read.post(
            object_dn=application_dn,
            attribute_name=ApplicationBaseAttributes.status
        )

        return response.values[0] if response.values else None

    def _is_in_error(self, application: 'Union[Config.Object, str]'):
        application_dn = self._get_dn(application)
        response = self._api.websdk.Config.Read.post(
            object_dn=application_dn,
            attribute_name=ApplicationBaseAttributes.in_error
        )

        return bool(response.values[0]) if response.values else False

    def wait_for_installation_to_complete(self, application: 'Union[Config.Object, str]', timeout: int = 60):
        """
        Waits for the application object's "Last Pushed On" attribute to be a date greater than
        or equal to the "Last Renewed On" date on the associated certificate. If the certificate
        has not been recently renewed and is simply being associated to the certificate, either
        clear the "Last Pushed On" date from the application object or use
        :meth:`pytpp.pytpp.features.certificate.Certificate.associate_application` with
        ``push_to_new=True``.

        Args:
            application: :ref:`config_object` or :ref:`dn` of the application object.
            timeout: Timeout in seconds.
        """
        application_dn = self._get_dn(application)
        certificate_dn = self._api.websdk.Config.Read.post(
            object_dn=application_dn,
            attribute_name=ApplicationBaseAttributes.certificate
        )
        if not certificate_dn.values:
            raise UnexpectedValue(
                f'There is no certificate associated to "{application_dn}" in TPP.'
            )
        response = self._api.websdk.Config.Read.post(
            object_dn=certificate_dn.values[0],
            attribute_name=X509CertificateAttributes.last_renewed_on
        )

        if not response.values:
            raise UnexpectedValue(
                f'Cannot validate that the certificate "{certificate_dn}" is installed on the application '
                f'"{application_dn}" as it seems that the certificate has never been renewed.'
            )
        certificate_last_renewed_time = from_date_string(response.values[0])

        def _certificate_is_installed():
            resp = self._api.websdk.Config.Read.post(
                object_dn=application_dn,
                attribute_name=ApplicationBaseAttributes.last_pushed_on
            )

            if not resp.values:
                return False
            application_last_pushed_on = from_date_string(resp.values[0])
            return application_last_pushed_on >= certificate_last_renewed_time

        stage = self._get_stage(application=application)
        with self._Timeout(timeout=timeout) as to:
            while not to.is_expired():
                if self._is_in_error(application=application):
                    break
                elif not stage:
                    if not _certificate_is_installed():
                        raise UnexpectedValue(
                            f'Expected a certificate to be installed on "{application_dn}", '
                            f'but the application is not in a processing status.'
                        )
                    return
                stage = self._get_stage(application=application)

        raise TimeoutError(
            f'Certificate installation failed on "{application_dn}".\n'
            f'Stage: {stage}\n'
            f'Status: {self.get_status(application=application)}'
        )


@feature('Adaptable Application')
class Adaptable(_ApplicationBase):
    def __init__(self, api):
        super().__init__(api=api, class_name=Classes.adaptable_app)

    @staticmethod
    def _calculate_hash(script_content: bytes):
        return b64encode(sha256(
            script_content.decode().encode('utf-32-le')
        ).hexdigest().encode()).decode()

    def create(self, name: 'str', device: 'Union[Config.Object, str]', policy_folder: 'Union[Config.Object, str]',
               powershell_script_name: 'str', powershell_script_content: 'bytes', locked: 'bool' = False,
               retry_after_scipt_hash_mismatch: 'bool' = None, description: 'str' = None,
               contacts: 'List[Union[Identity.Identity, str]]' = None,
               approvers: 'List[Union[Identity.Identity, str]]' = None,
               application_credential: 'Union[Config.Object, str]' = None,
               secondary_credential: 'Union[Config.Object, str]' = None,
               port: 'int' = None, private_key_credential: 'str' = None, log_debug: 'bool' = None,
               attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the application object.
            device: :ref:`config_object` or :ref:`dn` of the device object.
            policy_folder: :ref:`config_object` or :ref:`dn` of the folder to write the Adapatble PowerShell script policy.
            powershell_script_name: Name of the PowerShell script.
            powershell_script_content: Content of the PowerShell script in bytes. Use ``open(ps_script, 'rb')``.
            locked: Lock this script on the ``policy_folder``.
            retry_after_scipt_hash_mismatch: When the script is updated fix the related provisioning and discovery errors.
            description: Description for the application object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts for the application object.
            approvers: List of :ref:`identity_object` or :ref:`prefixed_name` as approvers for the application object.
            application_credential: :ref:`config_object` or :ref:`dn` of the application credential object.
            secondary_credential: :ref:`config_object` or :ref:`dn` of the supplemental application credential to pass to PowerShell script.
            port: Port number.
            private_key_credential: Optional private key credential.
            log_debug: Enable log debug.
            attributes: Additional attributes pertaining to the application object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the application.
        """

        # region Create The Policy Attributes
        policy_folder_dn = self._get_dn(obj=policy_folder)
        if retry_after_scipt_hash_mismatch is True:
            self._api.websdk.Config.WritePolicy.post(
                object_dn=policy_folder_dn,
                class_name=Classes.adaptable_app,
                attribute_name=AdaptableAppAttributes.script_hash_mismatch_error,
                values=[retry_after_scipt_hash_mismatch],
                locked=locked
            ).assert_valid_response()
        self._api.websdk.Config.WritePolicy.post(
            object_dn=policy_folder_dn,
            class_name=Classes.adaptable_app,
            attribute_name=AdaptableAppAttributes.powershell_script,
            values=[powershell_script_name],
            locked=locked
        ).assert_valid_response()
        vault_id = self._api.websdk.SecretStore.Add.post(
            base_64_data=self._calculate_hash(powershell_script_content),
            keyname=KeyNames.software_default,
            vault_type=VaultTypes.blob,
            namespace=Namespaces.config,
            owner=policy_folder_dn
        ).vault_id

        self._api.websdk.Config.WritePolicy.post(
            object_dn=policy_folder_dn,
            class_name=Classes.adaptable_app,
            attribute_name=AdaptableAppAttributes.powershell_script_hash_vault_id,
            values=[vault_id],
            locked=locked
        ).assert_valid_response()
        # endregion Create The Policy Attributes

        app_attrs = {
            ApplicationBaseAttributes.driver_name      : 'appadaptable',
            AdaptableAppAttributes.credential          : self._get_dn(
                application_credential) if application_credential else None,
            AdaptableAppAttributes.log_debug           : int(log_debug),
            AdaptableAppAttributes.pk_credential       : self._get_dn(
                private_key_credential) if private_key_credential else None,
            AdaptableAppAttributes.port                : port,
            AdaptableAppAttributes.secondary_credential: self._get_dn(
                secondary_credential) if secondary_credential else None,
        }
        if attributes:
            app_attrs.update(attributes)

        return self._create(
            name=name, device=device, contacts=contacts, approvers=approvers, description=description,
            attributes=app_attrs, get_if_already_exists=get_if_already_exists
        )


@feature('Amazon AWS')
class AmazonAWS(_ApplicationBase):
    def __init__(self, api):
        super().__init__(api=api, class_name=Classes.amazon_app)

    def create(self, name: str, device: 'Union[Config.Object, str]', aws_credential: 'Union[Config.Object, str]',
               description: 'str' = None,
               contacts: 'List[Union[Identity.Identity, str]]' = None,
               approvers: 'List[Union[Identity.Identity, str]]' = None,
               issued_by_aws: 'bool' = None, provision_to: 'int' = None, region: 'str' = None,
               iam_install_path: 'str' = None, replace_existing: 'bool' = None, binding_taget: 'Union[str, int]' = None,
               load_balancer_name: 'str' = None, load_balancer_port: 'int' = None, target_group: 'str' = None,
               create_listener: 'bool' = None, cloudfront_distribution_id: 'str' = None, attributes: dict = None,
               get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the application object.
            device: :ref:`config_object` or :ref:`dn` of the device object.
            aws_credential: :ref:`config_object` or :ref:`dn` of the AWS credential object.
            description: Description for the application object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts for the application object.
            approvers: List of :ref:`identity_object` or :ref:`prefixed_name` as approvers for the application object.
            issued_by_aws: The certificate should be issued by AWS Certificate Manager.
            provision_to: ACM or IAM.
            region: AWS region.
            iam_install_path: IAM path for certificate upload.
            replace_existing: Replace the existing store.
            binding_taget: Binding target.
            load_balancer_name: Name of the load balancer.
            load_balancer_port: Port of the load balancer.
            target_group: Default target group.
            create_listener: Create listener.
            cloudfront_distribution_id: CloudFront Distribution ID.
            attributes: Additional attributes pertaining to the application object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the application.
        """
        app_attrs = {
            ApplicationBaseAttributes.driver_name         : 'appamazon',
            AmazonAppAttributes.aws_credential_dn         : self._get_dn(aws_credential),
            AmazonAppAttributes.issued_by_aws             : {True: "1", False: "0"}.get(issued_by_aws),
            AmazonAppAttributes.provisioning_to           : provision_to,
            AmazonAppAttributes.load_balancer_region_code : region,
            AmazonAppAttributes.install_path              : iam_install_path,
            AmazonAppAttributes.replace_store             : {True: "1", False: "0"}.get(replace_existing),
            AmazonAppAttributes.binding_target            : binding_taget,
            AmazonAppAttributes.load_balancer_name        : load_balancer_name,
            AmazonAppAttributes.load_balancer_port        : load_balancer_port,
            AmazonAppAttributes.create_binding            : {True: "1", False: "0"}.get(create_listener),
            AmazonAppAttributes.target_group              : target_group,
            AmazonAppAttributes.cloudfront_distribution_id: cloudfront_distribution_id
        }
        if attributes:
            app_attrs.update(attributes)

        return self._create(
            name=name,
            device=device,
            approvers=approvers,
            contacts=contacts,
            description=description,
            attributes=app_attrs,
            get_if_already_exists=get_if_already_exists
        )


@feature('Apache')
class Apache(_ApplicationBase):
    def __init__(self, api):
        super().__init__(api=api, class_name=Classes.apache)

    def create(self, name: str, device: 'Union[Config.Object, str]', private_key_file: 'str', certificate_file: 'str',
               description: 'str' = None, contacts: 'List[Union[Identity.Identity, str]]' = None,
               approvers: 'List[Union[Identity.Identity, str]]' = None,
               application_credential: 'Union[Config.Object, str]' = None, port: 'int' = None,
               private_key_location: 'str' = None, client_tools_path: 'str' = None,
               protection_type: 'str' = None, softcard_identifier: 'str' = None, ocs_identifier: 'str' = None,
               partition_password_credential: 'Union[Config.Object, str]' = None,
               private_key_credential: 'Union[Config.Object, str]' = None,
               certificate_chain_file: 'str' = None, overwrite_existing_chain: 'bool' = None, owner: 'str' = None,
               owner_permissions: 'str' = None, group: 'str' = None, group_permissions: 'str' = None,
               attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the application object.
            device: :ref:`config_object` or :ref:`dn` of the device object.
            private_key_file: File location to place the private key file.
            certificate_file: File location to place the certificate file.
            description: Description for the application object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts for the application object.
            approvers: List of :ref:`identity_object` or :ref:`prefixed_name` as approvers for the application object.
            application_credential: :ref:`config_object` or :ref:`dn` of the application credential.
            port: Connection port.
            private_key_location: Remote generation setting.
            client_tools_path: HSM client tools path.
            protection_type: HSM protection type.
            softcard_identifier: Softcard label.
            ocs_identifier: OCS label.
            partition_password_credential: :ref:`config_object` or :ref:`dn` of OCS pin or softcard passphrase credential.
            private_key_credential: :ref:`config_object` or :ref:`dn` of private key credential.
            certificate_chain_file: File location to place the certificate chain file.
            overwrite_existing_chain: Overwirte the existing chain.
            owner: File user owner.
            owner_permissions: File permissions assigned to the user owner.
            group: File group owner.
            group_permissions: File permissions assigned to the group owner.
            attributes: Additional attributes pertaining to the application object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the application.
        """
        app_attrs = {
            ApplicationBaseAttributes.driver_name           : 'appapache',
            ApacheAttributes.credential                     : self._get_dn(
                application_credential) if application_credential else None,
            ApacheAttributes.port                           : port,
            ApacheAttributes.private_key_location           : private_key_location,
            ApacheAttributes.client_tools_path              : client_tools_path,
            ApacheAttributes.protection_type                : protection_type,
            ApacheAttributes.softcard_identifier            : softcard_identifier,
            ApacheAttributes.ocs_identifier                 : ocs_identifier,
            ApacheAttributes.partition_password_credential  : self._get_dn(
                partition_password_credential) if partition_password_credential else None,
            ApacheAttributes.private_key_file               : private_key_file,
            ApacheAttributes.private_key_password_credential: self._get_dn(
                private_key_credential) if private_key_credential else None,
            ApacheAttributes.certificate_file               : certificate_file,
            ApacheAttributes.certificate_chain_file         : certificate_chain_file,
            ApacheAttributes.overwrite_existing_chain       : {True: "1", False: "0"}.get(overwrite_existing_chain),
        }
        if owner or group:
            app_attrs.update({
                ApacheAttributes.file_permissions_enabled: "1",
                ApacheAttributes.file_owner_user         : owner,
                ApacheAttributes.file_permissions_user   : owner_permissions,
                ApacheAttributes.file_owner_group        : group,
                ApacheAttributes.file_permissions_group  : group_permissions,
            })

        if attributes:
            app_attrs.update(attributes)

        return self._create(
            name=name,
            device=device,
            approvers=approvers,
            contacts=contacts,
            description=description,
            attributes=app_attrs,
            get_if_already_exists=get_if_already_exists
        )


@feature('Azure Key Vault')
class AzureKeyVault(_ApplicationBase):
    def __init__(self, api):
        super().__init__(api=api, class_name=Classes.azure_key_vault)

    def create(self, name: str, device: 'Union[Config.Object, str]', application_id: 'str',
               certificate_credential: 'Union[Config.Object, str]', azure_key_vault_name: 'str',
               description: 'str' = None, contacts: 'List[Union[Identity.Identity, str]]' = None,
               approvers: 'List[Union[Identity.Identity, str]]' = None, certificate_name: 'str' = None,
               private_key_exportable: 'bool' = None, web_application_name: 'str' = None,
               create_new_binding: 'bool' = None, create_san_dns_bindings: 'bool' = None, ssl_type: 'int' = None,
               binding_hostnames: 'List[str]' = None, attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the application object.
            device: :ref:`config_object` or :ref:`dn` of the device object.
            application_id: Application ID.
            certificate_credential: :ref:`config_object` or :ref:`dn` of the certificate credential.
            azure_key_vault_name:  Azure Key Vault name.
            description: Description for the application object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts for the application object.
            approvers: List of :ref:`identity_object` or :ref:`prefixed_name` as approvers for the application object.
            certificate_name: Certificate name.
            private_key_exportable: Mark the private key as exportable.
            web_application_name: Name of the web application.
            create_new_binding: Create new certificate binding.
            create_san_dns_bindings: Create SAN DNS binding.
            ssl_type: SNI or IP Based.
            binding_hostnames: Binding hostnames.
            attributes: Additional attributes pertaining to the application object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the application.
        """
        app_attrs = {
            ApplicationBaseAttributes.driver_name         : 'appazurekeyvault',
            AzureKeyVaultAttributes.client_id             : application_id,
            AzureKeyVaultAttributes.certificate_credential: self._get_dn(certificate_credential),
            AzureKeyVaultAttributes.vault_name            : azure_key_vault_name,
            AzureKeyVaultAttributes.certificate_name      : certificate_name,
            AzureKeyVaultAttributes.non_exportable        : {True: "0", False: "1"}.get(private_key_exportable)
        }
        if web_application_name:
            app_attrs.update({
                AzureKeyVaultAttributes.update_web_app         : "0",
                AzureKeyVaultAttributes.web_app_name           : web_application_name,
                AzureKeyVaultAttributes.create_binding         : {True: "0", False: "1"}.get(create_new_binding),
                AzureKeyVaultAttributes.create_san_dns_bindings: {True: "0", False: "1"}.get(create_san_dns_bindings),
                AzureKeyVaultAttributes.binding_ssl_type       : ssl_type,
                AzureKeyVaultAttributes.binding_hostnames      : binding_hostnames
            })
        else:
            app_attrs.update({
                AzureKeyVaultAttributes.update_web_app: "1"
            })

        if attributes:
            app_attrs.update(attributes)

        return self._create(
            name=name,
            device=device,
            approvers=approvers,
            contacts=contacts,
            description=description,
            attributes=app_attrs,
            get_if_already_exists=get_if_already_exists
        )


@feature('Basic Application')
class Basic(_ApplicationBase):
    def __init__(self, api):
        super().__init__(api=api, class_name=Classes.basic)

    def create(self, name: str, device: 'Union[Config.Object, str]', description: 'str' = None,
               contacts: 'List[Union[Identity.Identity, str]]' = None,
               approvers: 'List[Union[Identity.Identity, str]]' = None, attributes: dict = None,
               get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the application object.
            device: :ref:`config_object` or :ref:`dn` of the device object.
            description: Description for the application object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts for the application object.
            approvers: List of :ref:`identity_object` or :ref:`prefixed_name` as approvers for the application object.
            attributes: Additional attributes pertaining to the application object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the application.
        """
        app_attrs = {
            ApplicationBaseAttributes.driver_name: 'appbasic'
        }
        if attributes:
            app_attrs.update(attributes)

        return self._create(
            name=name,
            device=device,
            approvers=approvers,
            contacts=contacts,
            description=description,
            attributes=app_attrs,
            get_if_already_exists=get_if_already_exists
        )

    def convert(self, basic_application: 'Union[Config.Object, str]', new_class_name: str, attributes: dict = None):
        """
        Converts the Basic application to another application class type. If attributes are given,
        then those attributes will apply to the new application once conversion is complete.

        Args:
            basic_application: :ref:`config_object` or :ref:`dn` of the Basic application.
            new_class_name: Application class name of the new application object.
            attributes: Attributes pertaining to the new application object.

        Returns:
            :ref:`config_object` of the application.
        """
        basic_application_dn = self._get_dn(basic_application)
        result = self._api.websdk.Config.MutateObject.post(
            object_dn=basic_application_dn,
            class_name=str(new_class_name)
        )
        result.assert_valid_response()

        if attributes:
            attributes = {k: ([str(v)] if not isinstance(v, list) else v) for k, v in attributes.items()}
            result = self._api.websdk.Config.Write.post(
                object_dn=basic_application_dn,
                attribute_data=self._name_value_list(attributes, keep_list_values=True)
            )
            result.assert_valid_response()

        new_object = self._api.websdk.Config.IsValid.post(object_dn=basic_application_dn).object
        if new_object.type_name != new_class_name:
            raise UnexpectedValue(
                f'Unable to convert Basic App "{basic_application_dn}" to {new_class_name}. '
                f'Its current class name is {new_object.type_name}.'
            )
        return new_object


@feature('Blue Coat SSLVA')
class BlueCoatSSLVA(_ApplicationBase):
    def __init__(self, api):
        super().__init__(api=api, class_name=Classes.bluecoat_sslva)

    def create(self, name: str, device: 'Union[Config.Object, str]', description: 'str' = None,
               contacts: 'List[Union[Identity.Identity, str]]' = None,
               approvers: 'List[Union[Identity.Identity, str]]' = None,
               application_credential: 'Union[Config.Object, str]' = None,
               port: 'int' = None, device_certificate: 'bool' = None, replace_existing: 'bool' = None,
               install_chain: 'bool' = None,
               create_lists: 'bool' = None, known_certificates_with_keys_lists: 'List[str]' = None,
               attributes: dict = None,
               get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the application object.
            device: :ref:`config_object` or :ref:`dn` of the device object.
            description: Description for the application object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts for the application object.
            approvers: List of :ref:`identity_object` or :ref:`prefixed_name` as approvers for the application object.
            application_credential: :ref:`config_object` or :ref:`dn` of the application credential.
            port: Connection port number.
            device_certificate: The key certificate is a device certificate.
            replace_existing: Replace existing certificate.
            install_chain: Install the certificate chain.
            create_lists: Create lists.
            known_certificates_with_keys_lists: Known certificates with keys list(s).
            attributes: Additional attributes pertaining to the application object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the application.
        """
        app_attrs = {
            ApplicationBaseAttributes.driver_name     : 'appBlueCoat',
            BlueCoatSSLVAAttributes.credential        : self._get_dn(application_credential) if application_credential else None,
            BlueCoatSSLVAAttributes.port              : port,
            BlueCoatSSLVAAttributes.device_certificate: {True: "1", False: "0"}.get(device_certificate),
            BlueCoatSSLVAAttributes.replace_store     : {True: "1", False: "0"}.get(replace_existing),
            BlueCoatSSLVAAttributes.certificate_only  : {True: "1", False: "0"}.get(install_chain),
            BlueCoatSSLVAAttributes.create_lists      : {True: "1", False: "0"}.get(create_lists),
            BlueCoatSSLVAAttributes.certificate_label : known_certificates_with_keys_lists
        }
        if attributes:
            app_attrs.update(attributes)

        return self._create(
            name=name,
            device=device,
            approvers=approvers,
            contacts=contacts,
            description=description,
            attributes=app_attrs,
            get_if_already_exists=get_if_already_exists
        )


@feature('CAPI')
class CAPI(_ApplicationBase):
    def __init__(self, api):
        super().__init__(api=api, class_name=Classes.capi)

    def create(self, name: str, device: 'Union[Config.Object, str]', description: 'str' = None,
               contacts: 'List[Union[Identity.Identity, str]]' = None,
               approvers: 'List[Union[Identity.Identity, str]]' = None,
               application_credential: 'Union[Config.Object, str]' = None, winrm_port: 'int' = None,
               private_key_location: 'str' = None, key_label: 'str' = None, friendly_name: 'str' = None,
               exportable: 'bool' = None,
               private_key_trustee: 'str' = None, web_site_name: 'str' = None, binding_ip_address: 'str' = None,
               binding_port: 'int' = None,
               binding_hostname: 'str' = None, create_binding: 'bool' = None, attributes: dict = None,
               get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the application object.
            device: :ref:`config_object` or :ref:`dn` of the device object.
            description: Description for the application object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts for the application object.
            approvers: List of :ref:`identity_object` or :ref:`prefixed_name` as approvers for the application object.
            application_credential: :ref:`config_object` or :ref:`dn` of the application credential.
            winrm_port: WinRM connection port.
            private_key_location: Remote key generation setting for generating the private key.
            key_label: Private key label. Only applies if using remote key generation.
            friendly_name: Friendly name.
            exportable: Private key is exportable.
            private_key_trustee: Private key trustee.
            web_site_name: Name of the web site.
            binding_ip_address: Binding IP Address.
            binding_port: Binding port.
            binding_hostname: Binding hostname.
            create_binding: Create the binding.
            attributes: Additional attributes pertaining to the application object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the application.
        """
        app_attrs = {
            ApplicationBaseAttributes.driver_name: 'appcapi',
            CAPIAttributes.credential            : self._get_dn(application_credential) if application_credential else None,
            CAPIAttributes.port                  : winrm_port,
            CAPIAttributes.private_key_location  : private_key_location,
            CAPIAttributes.private_key_label     : key_label,
            CAPIAttributes.friendly_name         : friendly_name,
            CAPIAttributes.non_exportable        : {True: "0", False: "1"}.get(exportable),
            CAPIAttributes.private_key_trustee   : private_key_trustee,
        }
        if web_site_name:
            app_attrs.update({
                CAPIAttributes.update_iis        : "1",
                CAPIAttributes.web_site_name     : web_site_name,
                CAPIAttributes.binding_ip_address: binding_ip_address,
                CAPIAttributes.binding_port      : binding_port,
                CAPIAttributes.hostname          : binding_hostname,
                CAPIAttributes.create_binding    : {True: "1", False: "0"}.get(create_binding),
            })
        else:
            app_attrs.update({
                CAPIAttributes.update_iis: "0"
            })

        if attributes:
            app_attrs.update(attributes)

        return self._create(
            name=name,
            device=device,
            approvers=approvers,
            contacts=contacts,
            description=description,
            attributes=app_attrs,
            get_if_already_exists=get_if_already_exists
        )


@feature('Citrix NetScaler')
class CitrixNetScaler(_ApplicationBase):
    def __init__(self, api):
        super().__init__(api=api, class_name=Classes.netscaler)

    def create(self, name: str, device: 'Union[Config.Object, str]', description: 'str' = None,
               contacts: 'List[Union[Identity.Identity, str]]' = None,
               approvers: 'List[Union[Identity.Identity, str]]' = None,
               application_credential: 'Union[Config.Object, str]' = None, port: 'int' = None,
               install_certificate_chain: 'bool' = None,
               use_fips: 'bool' = None, private_key_credential: 'Union[Config.Object, str]' = None,
               import_only: 'bool' = None,
               subfolder_relative_path: 'str' = None, certificate_binding: 'str' = None,
               virtual_server_name: 'str' = None,
               sni_certificate: 'bool' = None, attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the application object.
            device: :ref:`config_object` or :ref:`dn` of the device object.
            description: Description for the application object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts for the application object.
            approvers: List of :ref:`identity_object` or :ref:`prefixed_name` as approvers for the application object.
            application_credential: :ref:`config_object` or :ref:`dn` of the application credential.
            port: Connection port.
            install_certificate_chain: Install the certificate chain.
            use_fips: Use FIPS.
            private_key_credential: :ref:`config_object` or :ref:`dn` of the private key credential.
            import_only: Import only.
            subfolder_relative_path: Subfolder relative path.
            certificate_binding: Certificate binding.
            virtual_server_name: Virtual server name, service name, or service group name.
            sni_certificate: SNI certificate.
            attributes: Additional attributes pertaining to the application object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the application.
        """
        app_attrs = {
            ApplicationBaseAttributes.driver_name              : 'appnetscaler',
            NetScalerAttributes.credential                     : self._get_dn(application_credential) if application_credential else None,
            NetScalerAttributes.port                           : port,
            NetScalerAttributes.chain_cert                     : {True: "1", False: "0"}.get(install_certificate_chain),
            NetScalerAttributes.fips_key                       : {True: "1", False: "0"}.get(use_fips),
            NetScalerAttributes.private_key_password_credential: self._get_dn(private_key_credential) if private_key_credential else None,
            NetScalerAttributes.import_only                    : {True: "1", False: "0"}.get(import_only),
            NetScalerAttributes.install_path                   : subfolder_relative_path,
            NetScalerAttributes.ssl_object_type                : certificate_binding,
            NetScalerAttributes.virtual_server_name            : virtual_server_name,
            NetScalerAttributes.sni_certificate                : {True: "1", False: "0"}.get(sni_certificate),
        }

        if attributes:
            app_attrs.update(attributes)

        return self._create(
            name=name,
            device=device,
            approvers=approvers,
            contacts=contacts,
            description=description,
            attributes=app_attrs,
            get_if_already_exists=get_if_already_exists
        )


@feature('Connect:Direct')
class ConnectDirect(_ApplicationBase):
    def __init__(self, api):
        super().__init__(api=api, class_name=Classes.connectdirect)

    def create(self, name: str, device: 'Union[Config.Object, str]', description: 'str' = None,
               contacts: 'List[Union[Identity.Identity, str]]' = None,
               approvers: 'List[Union[Identity.Identity, str]]' = None,
               application_credential: 'Union[Config.Object, str]' = None, api_protocol: 'str' = None,
               port: 'int' = None, node_name: 'str' = None, install_chain: 'bool' = None,
               key_certificate_alias: 'str' = None, attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the application object.
            device: :ref:`config_object` or :ref:`dn` of the device object.
            description: Description for the application object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts for the application object.
            approvers: List of :ref:`identity_object` or :ref:`prefixed_name` as approvers for the application object.
            application_credential: :ref:`config_object` or :ref:`dn` of the application credential.
            api_protocol: API protocol.
            port: Connection port.
            node_name: Node name.
            install_chain: Install certificate chain.
            key_certificate_alias: Key certificate alias.
            attributes: Additional attributes pertaining to the application object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the application.
        """
        app_attrs = {
            ApplicationBaseAttributes.driver_name    : 'appConnectDirect',
            ConnectDirectAttributes.credential       : self._get_dn(application_credential) if application_credential else None,
            ConnectDirectAttributes.protocol         : api_protocol,
            ConnectDirectAttributes.port             : port,
            ConnectDirectAttributes.node_name        : node_name,
            ConnectDirectAttributes.certificate_only : {True: "1", False: "0"}.get(install_chain),
            ConnectDirectAttributes.certificate_label: key_certificate_alias
        }

        if attributes:
            app_attrs.update(attributes)

        return self._create(
            name=name,
            device=device,
            approvers=approvers,
            contacts=contacts,
            description=description,
            attributes=app_attrs,
            get_if_already_exists=get_if_already_exists
        )


@feature('F5 Authentication Bundle')
class F5AuthenticationBundle(_ApplicationBase):
    def __init__(self, api):
        super().__init__(api=api, class_name=Classes.f5_authentication_bundle)

    def create(self, name: str, device: 'Union[Config.Object, str]', bundle_file_name: str, description: 'str' = None,
               certifictes_to_use: 'List[Union[Config.Object, str]]' = None, attributes: dict = None,
               get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the application object.
            device: :ref:`config_object` or :ref:`dn` of the device object.
            bundle_file_name: Filename of the certificate bundle.
            description: Description for the application object.
            certifictes_to_use: List of  :ref:`config_object` or :ref:`dn` of the certificates to use.
            attributes: Additional attributes pertaining to the application object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the application.
        """
        app_attrs = {
            F5AuthenticationBundleAttributes.advanced_settings_bundle_name: bundle_file_name,
            F5AuthenticationBundleAttributes.description                  : description,
            F5AuthenticationBundleAttributes.certificates                 : [self._get_dn(c) for c in
                                                                             certifictes_to_use] if certifictes_to_use else None,
        }

        if attributes:
            app_attrs.update(attributes)

        return self._config_create(
            name=name,
            parent_folder_dn=self._get_dn(device),
            config_class=Classes.f5_authentication_bundle,
            attributes=app_attrs,
            get_if_already_exists=get_if_already_exists
        )


@feature('F5 LTM Advanced')
class F5LTMAdvanced(_ApplicationBase):
    def __init__(self, api):
        super().__init__(api=api, class_name=Classes.f5_ltm_advanced)

    def create(self, name: str, device: 'Union[Config.Object, str]', description: 'str' = None,
               contacts: 'List[Union[Identity.Identity, str]]' = None,
               approvers: 'List[Union[Identity.Identity, str]]' = None,
               application_credential: 'Union[Config.Object, str]' = None, https_port: 'int' = None,
               ssh_port: 'int' = None,
               device_certificate: 'bool' = None, provisioning_mode: 'int' = None,
               certificate_and_key_file: 'str' = None,
               private_key_credential: 'Union[Config.Object, str]' = None, force_profile_update: 'bool' = None,
               install_chain: 'bool' = None, bundle_certificate: 'bool' = None, overwrite_chain_file: 'bool' = None,
               ca_chain_file: 'str' = None, use_fips: 'bool' = None, overwrite_certificate_and_key: 'bool' = None,
               delete_previous_cert_and_key: 'bool' = None, provisioning_target: 'str' = None,
               config_sync: 'bool' = None,
               ssl_profile: 'str' = None, ssl_profile_type: 'str' = None, parent_ssl_profile: 'str' = None,
               ssl_partition: 'str' = None, sni_server_name: 'str' = None, sni_default: 'bool' = None,
               virtual_server: 'str' = None,
               virtual_server_partition: 'str' = None, use_advanced_settings: 'bool' = None,
               client_certificate_requirement: 'str' = None,
               server_certificate_requirement: 'str' = None, frequency: 'str' = None,
               chain_traversal_depth: 'int' = None,
               certificate_bundle: 'Union[Config.Object, str]' = None, authentication_name: 'str' = None,
               attributes: dict = None, get_if_already_exists: bool = True):
        """
        Creates a F5 LTM Advanced application object.

        Args:
            name: Name of the application object.
            device: :ref:`config_object` or :ref:`dn` of the device object.
            description: Description for the application object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts for the application object.
            approvers: List of :ref:`identity_object` or :ref:`prefixed_name` as approvers for the application object.
            application_credential: :ref:`config_object` or :ref:`dn` of the application credential.
            https_port: HTTPS port number.
            ssh_port: SSH port number.
            device_certificate: The certificate is a device certificate.
            provisioning_mode: Provisioning mode.
            certificate_and_key_file: Certificate and key file.
            private_key_credential: :ref:`config_object` or :ref:`dn` of the private key credential.
            force_profile_update: Force profile update.
            install_chain: Install the certificate chain.
            bundle_certificate: Bundle certificate.
            overwrite_chain_file: Overwrite the chain file.
            ca_chain_file: CA Chain file.
            use_fips: Use FIPS.
            overwrite_certificate_and_key: Overwrite certificate an key.
            delete_previous_cert_and_key: Delete old certificate and key.
            provisioning_target: Provisioning target for high availability.
            config_sync: Config sync.
            ssl_profile: SSL profile name.
            ssl_profile_type: SSL profile type.
            parent_ssl_profile: Parent SSL profile name.
            ssl_partition: SSL partition.
            sni_server_name: SNI server name.
            sni_default: SNI default.
            virtual_server: Virtual server name.
            virtual_server_partition: Virtual server partition.
            use_advanced_settings: Must be set if using advanced options.
            client_certificate_requirement: Require/request/ignore.
            server_certificate_requirement: Require/ignore.
            frequency: Once or Always.
            chain_traversal_depth: Chain traversal depth.
            certificate_bundle: :ref:`config_object` or :ref:`dn` of the certificate bundle.
            authentication_name: Authentication name.
            attributes: Additional attributes pertaining to the application object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the application.
        """
        app_attrs = {
            ApplicationBaseAttributes.driver_name                    : 'appf5ltmadvanced',
            F5LTMAdvancedAttributes.credential                       : self._get_dn(application_credential) if application_credential else None,
            F5LTMAdvancedAttributes.port                             : https_port,
            F5LTMAdvancedAttributes.ssh_port                         : ssh_port,
            F5LTMAdvancedAttributes.device_certificate               : {True: "1", False: "0"}.get(device_certificate),
            F5LTMAdvancedAttributes.use_basic_provisioning           : provisioning_mode,
            F5LTMAdvancedAttributes.certificate_name                 : certificate_and_key_file,
            F5LTMAdvancedAttributes.private_key_password_credential  : self._get_dn(private_key_credential) if private_key_credential else None,
            F5LTMAdvancedAttributes.force_profile_update             : {True: "1", False: "0"}.get(force_profile_update),
            F5LTMAdvancedAttributes.install_chain_file               : {True: "1", False: "0"}.get(install_chain),
            F5LTMAdvancedAttributes.bundle_certificate               : {True: "1", False: "0"}.get(bundle_certificate),
            F5LTMAdvancedAttributes.overwrite_existing_chain         : {True: "1", False: "0"}.get(overwrite_chain_file),
            F5LTMAdvancedAttributes.certificate_chain_name           : ca_chain_file,
            F5LTMAdvancedAttributes.fips_key                         : {True: "1", False: "0"}.get(use_fips),
            F5LTMAdvancedAttributes.overwrite_certificate            : {True: "1", False: "0"}.get(overwrite_certificate_and_key),
            F5LTMAdvancedAttributes.delete_previous_cert_and_key     : {True: "1", False: "0"}.get(delete_previous_cert_and_key),
            F5LTMAdvancedAttributes.provisioning_to                  : provisioning_target,
            F5LTMAdvancedAttributes.config_sync                      : {True: "1", False: "0"}.get(config_sync),
            F5LTMAdvancedAttributes.ssl_profile_name                 : ssl_profile,
            F5LTMAdvancedAttributes.ssl_profile_type                 : ssl_profile_type,
            F5LTMAdvancedAttributes.parent_ssl_profile_name          : parent_ssl_profile,
            F5LTMAdvancedAttributes.partition                        : ssl_partition,
            F5LTMAdvancedAttributes.sni_server_name                  : sni_server_name,
            F5LTMAdvancedAttributes.sni_default                      : {True: "1", False: "0"}.get(sni_default),
            F5LTMAdvancedAttributes.virtual_server_name              : virtual_server,
            F5LTMAdvancedAttributes.virtual_server_partition         : virtual_server_partition,
            F5LTMAdvancedAttributes.use_advanced_settings            : {True: "1", False: "0"}.get(use_advanced_settings),
            F5LTMAdvancedAttributes.client_authentication_certificate: client_certificate_requirement,
            F5LTMAdvancedAttributes.server_authentication_certificate: server_certificate_requirement,
            F5LTMAdvancedAttributes.authentication_frequency         : frequency,
            F5LTMAdvancedAttributes.chain_traversal_depth            : chain_traversal_depth,
            F5LTMAdvancedAttributes.bundle_certificate_collection    : self._get_dn(certificate_bundle) if certificate_bundle else None,
            F5LTMAdvancedAttributes.server_authentication_name       : authentication_name,
        }

        if attributes:
            app_attrs.update(attributes)

        return self._create(
            name=name,
            device=device,
            approvers=approvers,
            contacts=contacts,
            description=description,
            attributes=app_attrs,
            get_if_already_exists=get_if_already_exists
        )


@feature('Google Cloud Load Balancer')
class GoogleCloudLoadBalancer(_ApplicationBase):
    def __init__(self, api):
        super().__init__(api=api, class_name=Classes.google_cloud_app)

    def create(self, name: str, device: 'Union[Config.Object, str]', description: 'str' = None,
               contacts: 'List[Union[Identity.Identity, str]]' = None,
               approvers: 'List[Union[Identity.Identity, str]]' = None,
               google_credential: 'Union[Config.Object, str]' = None, target_proxy_type: 'str' = None,
               target_proxy_name: 'str' = None, target_resource: 'str' = None, attributes: dict = None,
               get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the application object.
            device: :ref:`config_object` or :ref:`dn` of the device object.
            description: Description for the application object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts for the application object.
            approvers: List of :ref:`identity_object` or :ref:`prefixed_name` as approvers for the application object.
            google_credential: :ref:`config_object` or :ref:`dn` of the application credential.
            target_proxy_type: Target proxy type.
            target_proxy_name: Target proxy name.
            target_resource: Target resource.
            attributes: Additional attributes pertaining to the application object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the application.
        """
        app_attrs = {
            ApplicationBaseAttributes.driver_name     : 'appgooglecloudloadbalancer',
            GoogleCloudAppAttributes.credential       : self._get_dn(google_credential) if google_credential else None,
            GoogleCloudAppAttributes.target_proxy_type: target_proxy_type,
            GoogleCloudAppAttributes.target_proxy_name: target_proxy_name,
            GoogleCloudAppAttributes.target_resource  : target_resource
        }

        if attributes:
            app_attrs.update(attributes)

        return self._create(
            name=name,
            device=device,
            approvers=approvers,
            contacts=contacts,
            description=description,
            attributes=app_attrs,
            get_if_already_exists=get_if_already_exists
        )


@feature('IBM DataPower')
class IBMDataPower(_ApplicationBase):
    def __init__(self, api):
        super().__init__(api=api, class_name=Classes.datapower)

    def create(self, name: str, device: 'Union[Config.Object, str]', description: 'str' = None,
               contacts: 'List[Union[Identity.Identity, str]]' = None,
               approvers: 'List[Union[Identity.Identity, str]]' = None,
               application_credential: 'Union[Config.Object, str]' = None, port: 'int' = None, xml_port: 'int' = None,
               basic_provisioning_mode: 'bool' = None, crypto_certificate: 'str' = None, crypto_key: 'str' = None,
               application_domain: 'str' = None, associate_to_profile: 'bool' = None, credential_type: 'str' = None,
               profile_type: 'str' = None, crypto_profile_name: 'str' = None, ssl_profile_name: 'str' = None,
               certificate_folder: 'str' = None, install_certificate_chain: 'bool' = None,
               private_key_password_credential: 'Union[Config.Object, str]' = None, attributes: dict = None,
               get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the application object.
            device: :ref:`config_object` or :ref:`dn` of the device object.
            description: Description for the application object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts for the application object.
            approvers: List of :ref:`identity_object` or :ref:`prefixed_name` as approvers for the application object.
            application_credential: :ref:`config_object` or :ref:`dn` of the application credential.
            port: Connection port.
            xml_port: XML connection port.
            basic_provisioning_mode: Use basic provisioning mode if ``True`` or advanced if ``False``.
            crypto_certificate: Crypto Certificate name.
            crypto_key: Crypto Private Key name.
            application_domain: Application domain.
            associate_to_profile: Associate to profile.
            credential_type: Credential type.
            profile_type: Profile type.
            crypto_profile_name: Crypto profile name.
            ssl_profile_name: SSL profile name.
            certificate_folder: Certificate folder.
            install_certificate_chain: Install the certficate chain.
            private_key_password_credential: :ref:`config_object` or :ref:`dn` of the private key password credential.
            attributes: Additional attributes pertaining to the application object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the application.
        """
        app_attrs = {
            ApplicationBaseAttributes.driver_name              : 'appdatapower',
            DataPowerAttributes.credential                     : self._get_dn(
                application_credential) if application_credential else None,
            DataPowerAttributes.port                           : port,
            DataPowerAttributes.xml_port                       : xml_port,
            DataPowerAttributes.use_basic_provisioning         : {True: "2", False: "1"}.get(basic_provisioning_mode),
            DataPowerAttributes.certificate_name               : crypto_certificate,
            DataPowerAttributes.private_key_name               : crypto_key,
            DataPowerAttributes.application_domain             : application_domain,
            DataPowerAttributes.associate_to_cp                : {True: "1", False: "2"}.get(associate_to_profile),
            DataPowerAttributes.credential_type                : credential_type,
            DataPowerAttributes.ssl_profile_type               : profile_type,
            DataPowerAttributes.crypto_profile                 : crypto_profile_name,
            DataPowerAttributes.ssl_proxy_profile              : ssl_profile_name,
            DataPowerAttributes.folder                         : certificate_folder,
            DataPowerAttributes.chain_cert                     : {True: "1", False: "0"}.get(install_certificate_chain),
            DataPowerAttributes.private_key_password_credential: self._get_dn(
                private_key_password_credential) if private_key_password_credential else None,
        }

        if attributes:
            app_attrs.update(attributes)

        return self._create(
            name=name,
            device=device,
            approvers=approvers,
            contacts=contacts,
            description=description,
            attributes=app_attrs,
            get_if_already_exists=get_if_already_exists
        )


@feature('IBM GSK')
class IBMGSK(_ApplicationBase):
    def __init__(self, api):
        super().__init__(api=api, class_name=Classes.gsk)

    def create(self, name: str, device: 'Union[Config.Object, str]', description: 'str' = None,
               contacts: 'List[Union[Identity.Identity, str]]' = None,
               approvers: 'List[Union[Identity.Identity, str]]' = None,
               application_credential: 'Union[Config.Object, str]' = None, port: 'int' = None, version: 'str' = None,
               gsk_utility_path: 'str' = None, java_home_path: 'str' = None, key_store_path: 'str' = None,
               key_store_credential: 'Union[Config.Object, str]' = None, create: 'bool' = None,
               replace_existing: 'bool' = None,
               certificate_label: 'str' = None, reuse_label: 'bool' = None, use_fips: 'bool' = None,
               password_validity: 'int' = None,
               stash_password: 'bool' = None, default_certificate: 'bool' = None, owner: 'str' = None,
               owner_permissions: 'str' = None, group: 'str' = None, group_permissions: 'str' = None,
               attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the application object.
            device: :ref:`config_object` or :ref:`dn` of the device object.
            description: Description for the application object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts for the application object.
            approvers: List of :ref:`identity_object` or :ref:`prefixed_name` as approvers for the application object.
            application_credential: :ref:`config_object` or :ref:`dn` of the application credential.
            port: Connection port.
            version: GSK version.
            gsk_utility_path: GSK utility path.
            java_home_path: JAVA_HOME path.
            key_store_path: Key store path.
            key_store_credential: :ref:`config_object` or :ref:`dn` of the key store credential.
            create: Create store.
            replace_existing: Replace existing store.
            certificate_label: Certificate label.
            reuse_label: Reuse certificate label.
            use_fips: Use FIPS.
            password_validity: Password validity in number of days.
            stash_password: Stash password.
            default_certificate: Default certificate.
            owner: File user owner.
            owner_permissions: File permissions assigned to the user owner.
            group: File group owner.
            group_permissions: File permissions assigned to the group owner.
            attributes: Additional attributes pertaining to the application object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the application.
        """
        app_attrs = {
            ApplicationBaseAttributes.driver_name: 'appgsk',
            GSKAttributes.credential             : self._get_dn(
                application_credential) if application_credential else None,
            GSKAttributes.port                   : port,
            GSKAttributes.version                : version,
            GSKAttributes.utility_path           : gsk_utility_path,
            GSKAttributes.java_home_path         : java_home_path,
            GSKAttributes.key_store              : key_store_path,
            GSKAttributes.key_store_credential   : self._get_dn(key_store_credential) if key_store_credential else None,
            GSKAttributes.create_store           : {True: "1", False: "0"}.get(create),
            GSKAttributes.replace_store          : {True: "1", False: "0"}.get(replace_existing),
            GSKAttributes.certificate_label      : certificate_label,
            GSKAttributes.recycle_alias          : {True: "1", False: "0"}.get(reuse_label),
            GSKAttributes.fips_key               : {True: "1", False: "0"}.get(use_fips),
            GSKAttributes.password_expire_days   : password_validity,
            GSKAttributes.stash_password         : {True: "1", False: "0"}.get(stash_password),
            GSKAttributes.default_cert           : {True: "1", False: "0"}.get(default_certificate),
        }
        if owner or group:
            app_attrs.update({
                GSKAttributes.file_permissions_enabled: "1",
                GSKAttributes.file_owner_user         : owner,
                GSKAttributes.file_permissions_user   : owner_permissions,
                GSKAttributes.file_owner_group        : group,
                GSKAttributes.file_permissions_group  : group_permissions
            })
        else:
            app_attrs.update({
                GSKAttributes.file_permissions_enabled: "0"
            })

        if attributes:
            app_attrs.update(attributes)

        return self._create(
            name=name,
            device=device,
            approvers=approvers,
            contacts=contacts,
            description=description,
            attributes=app_attrs,
            get_if_already_exists=get_if_already_exists
        )


@feature('Imperva MX')
class ImpervaMX(_ApplicationBase):
    def __init__(self, api):
        super().__init__(api=api, class_name=Classes.imperva_mx)

    def create(self, name: str, device: 'Union[Config.Object, str]', description: 'str' = None,
               contacts: 'List[Union[Identity.Identity, str]]' = None,
               approvers: 'List[Union[Identity.Identity, str]]' = None,
               application_credential: 'Union[Config.Object, str]' = None, port: 'int' = None,
               user_credential: 'Union[Config.Object, str]' = None, ssl_key_tool_path: 'str' = None, site: 'str' = None,
               server_group: 'str' = None, service: 'str' = None, attributes: dict = None,
               get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the application object.
            device: :ref:`config_object` or :ref:`dn` of the device object.
            description: Description for the application object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts for the application object.
            approvers: List of :ref:`identity_object` or :ref:`prefixed_name` as approvers for the application object.
            application_credential: :ref:`config_object` or :ref:`dn` of the application credential.
            port: Connection port.
            user_credential: :ref:`config_object` or :ref:`dn` of the user credential.
            ssl_key_tool_path: SSL key tool path.
            site: Site.
            server_group: Server group.
            service: Service.
            attributes: Additional attributes pertaining to the application object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the application.
        """
        app_attrs = {
            ApplicationBaseAttributes.driver_name  : 'appimpervamx',
            ImpervaMXAttributes.credential         : self._get_dn(
                application_credential) if application_credential else None,
            ImpervaMXAttributes.port               : port,
            ImpervaMXAttributes.username_credential: self._get_dn(user_credential) if user_credential else None,
            ImpervaMXAttributes.utility_path       : ssl_key_tool_path,
            ImpervaMXAttributes.site               : site,
            ImpervaMXAttributes.server_group       : server_group,
            ImpervaMXAttributes.service            : service
        }

        if attributes:
            app_attrs.update(attributes)

        return self._create(
            name=name,
            device=device,
            approvers=approvers,
            contacts=contacts,
            description=description,
            attributes=app_attrs,
            get_if_already_exists=get_if_already_exists
        )


@feature('JKS')
class JKS(_ApplicationBase):
    def __init__(self, api):
        super().__init__(api=api, class_name=Classes.jks)

    def create(self, name: str, device: 'Union[Config.Object, str]', description: 'str' = None,
               contacts: 'List[Union[Identity.Identity, str]]' = None,
               approvers: 'List[Union[Identity.Identity, str]]' = None,
               application_credential: 'Union[Config.Object, str]' = None, port: 'int' = None,
               private_key_location: 'str' = None,
               slot_number: 'int' = None, java_vendor: 'str' = None, protection_type: 'str' = None,
               softcard_identifier: 'str' = None,
               keytool_path: 'str' = None, version: 'str' = None, store_type: 'str' = None, keystore_path: 'str' = None,
               keystore_credential: 'Union[Config.Object, str]' = None,
               private_key_credential: 'Union[Config.Object, str]' = None,
               create: 'bool' = None, replace_existing: 'bool' = None, certificate_alias: 'str' = None,
               reuse_alias: 'bool' = None,
               owner: 'str' = None, owner_permissions: 'str' = None, group: 'str' = None,
               group_permissions: 'str' = None,
               attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the application object.
            device: :ref:`config_object` or :ref:`dn` of the device object.
            description: Description for the application object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts for the application object.
            approvers: List of :ref:`identity_object` or :ref:`prefixed_name` as approvers for the application object.
            application_credential: :ref:`config_object` or :ref:`dn` of the application credential.
            port: Connection port.
            private_key_location: Remote generation setting.
            slot_number: Slot number.
            java_vendor: Java vendor.
            protection_type: Remote generation protection type.
            softcard_identifier: Softcard identifier.
            keytool_path: Keytool path.
            version: Java version.
            store_type: Store type.
            keystore_path: Key store path.
            keystore_credential: :ref:`config_object` or :ref:`dn` of the key store credential.
            private_key_credential: :ref:`config_object` or :ref:`dn` of the private key credential.
            create: Create store.
            replace_existing: Replace existing store.
            certificate_alias: Certificate alias.
            reuse_alias: Resuse certificate alias.
            owner: File user owner.
            owner_permissions: File permissions assigned to the user owner.
            group: File group owner.
            group_permissions: File permissions assigned to the group owner.
            attributes: Additional attributes pertaining to the application object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the application.
        """
        app_attrs = {
            ApplicationBaseAttributes.driver_name        : 'appjks',
            JKSAttributes.credential                     : self._get_dn(
                application_credential) if application_credential else None,
            JKSAttributes.port                           : port,
            JKSAttributes.private_key_location           : private_key_location,
            JKSAttributes.slot_number                    : slot_number,
            JKSAttributes.java_vendor                    : java_vendor,
            JKSAttributes.protection_type                : protection_type,
            JKSAttributes.softcard_identifier            : softcard_identifier,
            JKSAttributes.keytool_path                   : keytool_path,
            JKSAttributes.version                        : version,
            JKSAttributes.store_type                     : store_type,
            JKSAttributes.key_store                      : keystore_path,
            JKSAttributes.key_store_credential           : self._get_dn(
                keystore_credential) if keystore_credential else None,
            JKSAttributes.private_key_password_credential: self._get_dn(
                private_key_credential) if private_key_credential else None,
            JKSAttributes.create_store                   : {True: "1", False: "0"}.get(create),
            JKSAttributes.replace_store                  : {True: "1", False: "0"}.get(replace_existing),
            JKSAttributes.certificate_label              : certificate_alias,
            JKSAttributes.recycle_alias                  : {True: "1", False: "0"}.get(reuse_alias)
        }
        if owner or group:
            app_attrs.update({
                JKSAttributes.file_permissions_enabled: "1",
                JKSAttributes.file_owner_user         : owner,
                JKSAttributes.file_permissions_user   : owner_permissions,
                JKSAttributes.file_owner_group        : group,
                JKSAttributes.file_permissions_group  : group_permissions
            })
        else:
            app_attrs.update({
                JKSAttributes.file_permissions_enabled: "0"
            })

        if attributes:
            app_attrs.update(attributes)

        return self._create(
            name=name,
            device=device,
            approvers=approvers,
            contacts=contacts,
            description=description,
            attributes=app_attrs,
            get_if_already_exists=get_if_already_exists
        )


@feature('Oracle iPlanet')
class OracleIPlanet(_ApplicationBase):
    def __init__(self, api):
        super().__init__(api=api, class_name=Classes.iplanet)

    def create(self, name: str, device: 'Union[Config.Object, str]', description: 'str' = None,
               contacts: 'List[Union[Identity.Identity, str]]' = None,
               approvers: 'List[Union[Identity.Identity, str]]' = None,
               application_credential: 'Union[Config.Object, str]' = None, port: 'int' = None,
               certificate_database_type: 'str' = None, certificate_database_path: 'str' = None,
               certificate_database_credential: 'Union[Config.Object, str]' = None,
               certificate_database_prefix: 'str' = None,
               create: 'bool' = None, replace_existing: 'bool' = None, certutil_path: 'str' = None,
               pk12util_path: 'str' = None,
               certificate_alias: 'str' = None, owner: 'str' = None, owner_permissions: 'str' = None,
               group: 'str' = None,
               group_permissions: 'str' = None, attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the application object.
            device: :ref:`config_object` or :ref:`dn` of the device object.
            description: Description for the application object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts for the application object.
            approvers: List of :ref:`identity_object` or :ref:`prefixed_name` as approvers for the application object.
            application_credential: :ref:`config_object` or :ref:`dn` of the application credential.
            port: Connection port.
            certificate_database_type: Certificate database type.
            certificate_database_path: Certificate database path.
            certificate_database_credential: :ref:`config_object` or :ref:`dn` of the ccertificate database credentiaL.
            certificate_database_prefix: Certificate database prefix.
            create: Create store.
            replace_existing: Replace existing store.
            certutil_path: Certutil path.
            pk12util_path: Pk12util path.
            certificate_alias: Certificate alias.
            owner: File user owner.
            owner_permissions: File permissions assigned to the user owner.
            group: File group owner.
            group_permissions: File permissions assigned to the group owner.
            attributes: Additional attributes pertaining to the application object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the application.
        """
        app_attrs = {
            ApplicationBaseAttributes.driver_name : 'appiplanet',
            iPlanetAttributes.credential          : self._get_dn(
                application_credential) if application_credential else None,
            iPlanetAttributes.port                : port,
            iPlanetAttributes.database_type       : certificate_database_type,
            iPlanetAttributes.key_store           : certificate_database_path,
            iPlanetAttributes.key_store_credential: self._get_dn(
                certificate_database_credential) if certificate_database_credential else None,
            iPlanetAttributes.database_prefix     : certificate_database_prefix,
            iPlanetAttributes.create_store        : {True: "1", False: "0"}.get(create),
            iPlanetAttributes.replace_store       : {True: "1", False: "0"}.get(replace_existing),
            iPlanetAttributes.certutil_path       : certutil_path,
            iPlanetAttributes.pk12util_path       : pk12util_path,
            iPlanetAttributes.alias               : certificate_alias,
        }
        if owner or group:
            app_attrs.update({
                iPlanetAttributes.file_permissions_enabled: "1",
                iPlanetAttributes.file_owner_user         : owner,
                iPlanetAttributes.file_permissions_user   : owner_permissions,
                iPlanetAttributes.file_owner_group        : group,
                iPlanetAttributes.file_permissions_group  : group_permissions
            })
        else:
            app_attrs.update({
                iPlanetAttributes.file_permissions_enabled: "0"
            })

        if attributes:
            app_attrs.update(attributes)

        return self._create(
            name=name,
            device=device,
            approvers=approvers,
            contacts=contacts,
            description=description,
            attributes=app_attrs,
            get_if_already_exists=get_if_already_exists
        )


@feature('Palo Alto Network FW')
class PaloAltoNetworkFW(_ApplicationBase):
    def __init__(self, api):
        super().__init__(api=api, class_name=Classes.palo_alto_network_fw)

    def create(self, name: str, device: 'Union[Config.Object, str]', description: 'str' = None,
               contacts: 'List[Union[Identity.Identity, str]]' = None,
               approvers: 'List[Union[Identity.Identity, str]]' = None,
               application_credential: 'Union[Config.Object, str]' = None, port: 'int' = None,
               provision_certificate_only: 'bool' = None, private_key_password: 'Union[Config.Object, str]' = None,
               install_chain: 'bool' = None, replace_certificate: 'bool' = None,
               decryption_policy_rule_name: 'str' = None,
               create_decryption_policy_rule: 'bool' = None, decryption_profile_name: 'str' = None,
               destination_addresses: 'List[str]' = None, lock_config: 'bool' = None,
               attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the application object.
            device: :ref:`config_object` or :ref:`dn` of the device object.
            description: Description for the application object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts for the application object.
            approvers: List of :ref:`identity_object` or :ref:`prefixed_name` as approvers for the application object.
            application_credential: :ref:`config_object` or :ref:`dn` of the application credential.
            port: Connection port.
            provision_certificate_only: If ``False``, also provision the private key.
            private_key_password: :ref:`config_object` or :ref:`dn` of the private key password.
            install_chain: Insatll the certifcate chain.
            replace_certificate: Replace certificate.
            decryption_policy_rule_name: Decryption policy rule name.
            create_decryption_policy_rule: Create decryption policy rule.
            decryption_profile_name: Decryption profile name.
            destination_addresses: Destination addresses.
            lock_config: Lock config.
            attributes: Additional attributes pertaining to the application object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the application.
        """
        app_attrs = {
            ApplicationBaseAttributes.driver_name                      : 'appPaloAlto',
            PaloAltoNetworkFWAttributes.credential                     : self._get_dn(
                application_credential) if application_credential else None,
            PaloAltoNetworkFWAttributes.port                           : port,
            PaloAltoNetworkFWAttributes.certificate_only               : {True: "1", False: "2"}.get(
                provision_certificate_only),
            PaloAltoNetworkFWAttributes.private_key_password_credential: self._get_dn(
                private_key_password) if private_key_password else None,
            PaloAltoNetworkFWAttributes.chain_cert                     : {True: "1", False: "0"}.get(install_chain),
            PaloAltoNetworkFWAttributes.replace_store                  : {True: "1", False: "0"}.get(
                replace_certificate),
            PaloAltoNetworkFWAttributes.decryption_policy              : decryption_policy_rule_name,
            PaloAltoNetworkFWAttributes.create_decryption_policy       : {True: "1", False: "0"}.get(
                create_decryption_policy_rule),
            PaloAltoNetworkFWAttributes.decryption_profile             : decryption_profile_name,
            PaloAltoNetworkFWAttributes.decryption_destinations        : destination_addresses,
            PaloAltoNetworkFWAttributes.lock_config                    : {True: "1", False: "0"}.get(lock_config)
        }

        if attributes:
            app_attrs.update(attributes)

        return self._create(
            name=name,
            device=device,
            approvers=approvers,
            contacts=contacts,
            description=description,
            attributes=app_attrs,
            get_if_already_exists=get_if_already_exists
        )


@feature('PEM')
class PEM(_ApplicationBase):
    def __init__(self, api):
        super().__init__(api=api, class_name=Classes.pem)

    def create(self, name: str, device: 'Union[Config.Object, str]', private_key_file: 'str', certificate_file: 'str',
               description: 'str' = None, contacts: 'List[Union[Identity.Identity, str]]' = None,
               approvers: 'List[Union[Identity.Identity, str]]' = None,
               application_credential: 'Union[Config.Object, str]' = None, port: 'int' = None,
               private_key_credential: 'Union[Config.Object, str]' = None, certificate_chain_file: 'str' = None,
               overwrite_existing_chain: 'bool' = None, owner: 'str' = None, owner_permissions: 'str' = None,
               group: 'str' = None, group_permissions: 'str' = None, attributes: dict = None,
               get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the application object.
            device: :ref:`config_object` or :ref:`dn` of the device object.
            private_key_file: File location to place the private key file.
            certificate_file: File location to place the certificate file.
            description: Description for the application object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts for the application object.
            approvers: List of :ref:`identity_object` or :ref:`prefixed_name` as approvers for the application object.
            application_credential: :ref:`config_object` or :ref:`dn` of the application credential.
            port: Connection port.
            private_key_credential: :ref:`config_object` or :ref:`dn` of private key credential.
            certificate_chain_file: File location to place the certificate chain file.
            overwrite_existing_chain: Overwirte the existing chain.
            owner: File user owner.
            owner_permissions: File permissions assigned to the user owner.
            group: File group owner.
            group_permissions: File permissions assigned to the group owner.
            attributes: Additional attributes pertaining to the application object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the application.
        """
        app_attrs = {
            ApplicationBaseAttributes.driver_name           : 'appPem',
            ApacheAttributes.credential                     : self._get_dn(
                application_credential) if application_credential else None,
            ApacheAttributes.port                           : port,
            ApacheAttributes.private_key_file               : private_key_file,
            ApacheAttributes.private_key_password_credential: self._get_dn(
                private_key_credential) if private_key_credential else None,
            ApacheAttributes.certificate_file               : certificate_file,
            ApacheAttributes.certificate_chain_file         : certificate_chain_file,
            ApacheAttributes.overwrite_existing_chain       : {True: "1", False: "0"}.get(overwrite_existing_chain),
        }
        if owner or group:
            app_attrs.update({
                ApacheAttributes.file_permissions_enabled: "1",
                ApacheAttributes.file_owner_user         : owner,
                ApacheAttributes.file_permissions_user   : owner_permissions,
                ApacheAttributes.file_owner_group        : group,
                ApacheAttributes.file_permissions_group  : group_permissions,
            })

        if attributes:
            app_attrs.update(attributes)

        return self._create(
            name=name,
            device=device,
            approvers=approvers,
            contacts=contacts,
            description=description,
            attributes=app_attrs,
            get_if_already_exists=get_if_already_exists
        )


@feature('PKCS11')
class PKCS11(_ApplicationBase):
    """
    .. note::
        The PKCS #11 VSE must be installed in order to use this driver.
    """
    def __init__(self, api):
        super().__init__(api=api, class_name=Classes.pkcs11)

    def create(self, name: str, device: 'Union[Config.Object, str]', connection_method: 'str', port: 'int',
               protection_type: 'str', token_identifier: 'str', token_pin: 'Union[Config.Object, str]',
               label_format: 'str',
               use_case: 'str', import_certificate_into_hsm: 'str', distribution_directory: 'str', cryptoki_file: 'str',
               openssl_config_file: 'str', description: 'str' = None,
               contacts: 'List[Union[Identity.Identity, str]]' = None,
               approvers: 'List[Union[Identity.Identity, str]]' = None,
               application_credential: 'Union[Config.Object, str]' = None,
               reverse_subject_dn: 'bool' = None, embed_sans_in_csr: 'bool' = None, requested_label: 'str' = None,
               client_tools_directory: 'str' = None, openssl_directory: 'str' = None, attributes: dict = None,
               get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the application object.
            device: :ref:`config_object` or :ref:`dn` of the device object.
            description: Description for the application object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts for the application object.
            approvers: List of :ref:`identity_object` or :ref:`prefixed_name` as approvers for the application object.
            application_credential: :ref:`config_object` or :ref:`dn` of the application credential.
            connection_method: WinRM or SSH.
            port: Connection port.
            protection_type: HSM protection type.
            token_identifier: Token/Slot identifier.
            token_pin: :ref:`config_object` or :ref:`dn` of the token/slot pin, or passphrase, credential.
            label_format: Label format.
            use_case: Use case.
            import_certificate_into_hsm: Where to import the certificate.
            distribution_directory: Distribution directory.
            cryptoki_file: Cryptoki file path.
            openssl_config_file: OpenSSL config file path.
            reverse_subject_dn: Reverse the subject :ref:`dn`.
            embed_sans_in_csr: Embed SANs in CSR.
            requested_label: Requested key label.
            client_tools_directory: Client tools directory.
            openssl_directory: OpenSSL directory.
            attributes: Additional attributes pertaining to the application object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the application.
        """
        app_attrs = {
            ApplicationBaseAttributes.driver_name     : 'apppkcs11',
            PKCS11Attributes.credential               : self._get_dn(
                application_credential) if application_credential else None,
            PKCS11Attributes.connection_method        : connection_method,
            PKCS11Attributes.port                     : port,
            PKCS11Attributes.hsm_protection_type      : protection_type,
            PKCS11Attributes.hsm_token_label          : token_identifier,
            PKCS11Attributes.hsm_token_password       : self._get_dn(token_pin),
            PKCS11Attributes.hsm_cka_label_format     : label_format,
            PKCS11Attributes.hsm_requested_cka_label  : requested_label,
            PKCS11Attributes.hsm_requested_usecase    : use_case,
            PKCS11Attributes.hsm_import_certificate   : import_certificate_into_hsm,
            PKCS11Attributes.hsm_certificate_directory: distribution_directory,
            PKCS11Attributes.hsm_cryptoki_file        : cryptoki_file,
            PKCS11Attributes.hsm_client_tool_path     : client_tools_directory,
            PKCS11Attributes.hsm_openssl_config_file  : openssl_config_file,
            PKCS11Attributes.hsm_reverse_subject_dn   : {True: "Yes", False: "No"}.get(reverse_subject_dn),
            PKCS11Attributes.hsm_embed_sans_in_csr    : {True: "Yes", False: "No"}.get(embed_sans_in_csr),
        }
        if openssl_directory:
            app_attrs.update({
                PKCS11Attributes.hsm_openssl_type: "Custom OpenSSL Directory",
                PKCS11Attributes.hsm_openssl_path: openssl_directory
            })

        if attributes:
            app_attrs.update(attributes)

        return self._create(
            name=name,
            device=device,
            approvers=approvers,
            contacts=contacts,
            description=description,
            attributes=app_attrs,
            get_if_already_exists=get_if_already_exists
        )


@feature('PKCS #12')
class PKCS12(_ApplicationBase):
    def __init__(self, api):
        super().__init__(api=api, class_name=Classes.pkcs_12)

    def create(self, name: str, device: 'Union[Config.Object, str]', pkcs12_file: 'str', description: 'str' = None,
               contacts: 'List[Union[Identity.Identity, str]]' = None,
               approvers: 'List[Union[Identity.Identity, str]]' = None,
               application_credential: 'Union[Config.Object, str]' = None, port: 'int' = None,
               private_key_credential: 'Union[Config.Object, str]' = None, friendly_name: 'str' = None,
               certificate_chain_file: 'str' = None, create: 'bool' = None, replace_existing: 'bool' = None,
               reuse_friendly_name: 'bool' = None, owner: 'str' = None,
               owner_permissions: 'str' = None, group: 'str' = None, group_permissions: 'str' = None,
               attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the application object.
            device: :ref:`config_object` or :ref:`dn` of the device object.
            pkcs12_file: PKCS #12 file.
            description: Description for the application object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts for the application object.
            approvers: List of :ref:`identity_object` or :ref:`prefixed_name` as approvers for the application object.
            application_credential: :ref:`config_object` or :ref:`dn` of the application credential.
            port: Connection port.
            private_key_credential: :ref:`config_object` or :ref:`dn` of the private key credential.
            friendly_name: Friendly name.
            certificate_chain_file: Certificate chain file.
            create: Create store.
            replace_existing: Replace existing store.
            reuse_friendly_name: Reuse the friendly name.
            owner: File user owner.
            owner_permissions: File permissions assigned to the user owner.
            group: File group owner.
            group_permissions: File permissions assigned to the group owner.
            attributes: Additional attributes pertaining to the application object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the application.
        """
        app_attrs = {
            ApplicationBaseAttributes.driver_name           : 'apppkcs12',
            PKCS12Attributes.credential                     : self._get_dn(
                application_credential) if application_credential else None,
            PKCS12Attributes.port                           : port,
            PKCS12Attributes.certificate_file               : pkcs12_file,
            PKCS12Attributes.private_key_password_credential: self._get_dn(
                private_key_credential) if private_key_credential else None,
            PKCS12Attributes.friendly_name                  : friendly_name,
            PKCS12Attributes.create_store                   : {True: "1", False: "0"}.get(create),
            PKCS12Attributes.replace_store                  : {True: "1", False: "0"}.get(replace_existing),
            PKCS12Attributes.recycle_alias                  : {True: "1", False: "0"}.get(reuse_friendly_name)
        }

        if certificate_chain_file:
            app_attrs.update({
                PKCS12Attributes.bundle_certificate    : "0",
                PKCS12Attributes.certificate_chain_file: certificate_chain_file
            })
        else:
            app_attrs.update({
                PKCS12Attributes.bundle_certificate: "1"
            })

        if owner or group:
            app_attrs.update({
                ApacheAttributes.file_permissions_enabled: "1",
                ApacheAttributes.file_owner_user         : owner,
                ApacheAttributes.file_permissions_user   : owner_permissions,
                ApacheAttributes.file_owner_group        : group,
                ApacheAttributes.file_permissions_group  : group_permissions,
            })

        if attributes:
            app_attrs.update(attributes)

        return self._create(
            name=name,
            device=device,
            approvers=approvers,
            contacts=contacts,
            description=description,
            attributes=app_attrs,
            get_if_already_exists=get_if_already_exists
        )


@feature('Riverbed Steelhead')
class RiverbedSteelHead(_ApplicationBase):
    def __init__(self, api):
        super().__init__(api=api, class_name=Classes.riverbed_steelhead)

    def create(self, name: str, device: 'Union[Config.Object, str]', description: 'str' = None,
               contacts: 'List[Union[Identity.Identity, str]]' = None,
               approvers: 'List[Union[Identity.Identity, str]]' = None,
               certificate_type: 'str' = None, replace_existing: 'bool' = None,
               install_chain_certificates: 'bool' = None,
               attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the application object.
            device: :ref:`config_object` or :ref:`dn` of the device object.
            description: Description for the application object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts for the application object.
            approvers: List of :ref:`identity_object` or :ref:`prefixed_name` as approvers for the application object.
            certificate_type: Certificate type.
            replace_existing: Replace existing certificate.
            install_chain_certificates: Install the chain certificates.
            attributes: Additional attributes pertaining to the application object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the application.
        """
        app_attrs = {
            ApplicationBaseAttributes.driver_name       : 'appriverbedsteelhead',
            RiverbedSteelHeadAttributes.certificate_type: certificate_type,
            RiverbedSteelHeadAttributes.replace_existing: {True: "1", False: "0"}.get(replace_existing),
            RiverbedSteelHeadAttributes.install_chain   : {True: "1", False: "0"}.get(install_chain_certificates)
        }

        if attributes:
            app_attrs.update(attributes)

        return self._create(
            name=name,
            device=device,
            approvers=approvers,
            contacts=contacts,
            description=description,
            attributes=app_attrs,
            get_if_already_exists=get_if_already_exists
        )


@feature('Tealeaf PCA')
class TealeafPCA(_ApplicationBase):
    def __init__(self, api):
        super().__init__(api=api, class_name=Classes.tealeaf_pca)

    def create(self, name: str, device: 'Union[Config.Object, str]', description: 'str' = None,
               contacts: 'List[Union[Identity.Identity, str]]' = None,
               approvers: 'List[Union[Identity.Identity, str]]' = None,
               application_credential: 'Union[Config.Object, str]' = None, port: 'int' = None,
               passive_capture_setup_path: 'str' = None, attributes: dict = None, get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the application object.
            device: :ref:`config_object` or :ref:`dn` of the device object.
            description: Description for the application object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts for the application object.
            approvers: List of :ref:`identity_object` or :ref:`prefixed_name` as approvers for the application object.
            application_credential: :ref:`config_object` or :ref:`dn` of the application credential.
            port: Connection port.
            passive_capture_setup_path: Passive capture setup path.
            attributes: Additional attributes pertaining to the application object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the application.
        """
        app_attrs = {
            ApplicationBaseAttributes.driver_name: 'apptealeafpca',
            TealeafPCAAttributes.credential      : self._get_dn(
                application_credential) if application_credential else None,
            TealeafPCAAttributes.port            : port,
            TealeafPCAAttributes.install_path    : passive_capture_setup_path
        }

        if attributes:
            app_attrs.update(attributes)

        return self._create(
            name=name,
            device=device,
            approvers=approvers,
            contacts=contacts,
            description=description,
            attributes=app_attrs,
            get_if_already_exists=get_if_already_exists
        )


@feature('VAM nShield')
class VAMnShield(_ApplicationBase):
    def __init__(self, api):
        super().__init__(api=api, class_name=Classes.vam_nshield)

    def create(self, name: str, device: 'Union[Config.Object, str]', description: 'str' = None,
               contacts: 'List[Union[Identity.Identity, str]]' = None,
               approvers: 'List[Union[Identity.Identity, str]]' = None,
               application_credential: 'Union[Config.Object, str]' = None, port: 'int' = None,
               nshield_setup_path: 'str' = None,
               module_id: 'int' = None, restart_device: 'bool' = None, attributes: dict = None,
               get_if_already_exists: bool = True):
        """
        Args:
            name: Name of the application object.
            device: :ref:`config_object` or :ref:`dn` of the device object.
            description: Description for the application object.
            contacts: List of :ref:`identity_object` or :ref:`prefixed_name` as contacts for the application object.
            approvers: List of :ref:`identity_object` or :ref:`prefixed_name` as approvers for the application object.
            application_credential: :ref:`config_object` or :ref:`dn` of the application credential.
            port: Connection port.
            nshield_setup_path: nShield setup path.
            module_id: Module ID.
            restart_device: Restart the device.
            attributes: Additional attributes pertaining to the application object.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the application.
        """
        app_attrs = {
            ApplicationBaseAttributes.driver_name   : 'appvamnshield',
            VAMnShieldAttributes.credential         : self._get_dn(
                application_credential) if application_credential else None,
            VAMnShieldAttributes.port               : port,
            VAMnShieldAttributes.install_path       : nshield_setup_path,
            VAMnShieldAttributes.module_id          : module_id,
            VAMnShieldAttributes.restart_application: restart_device
        }

        if attributes:
            app_attrs.update(attributes)

        return self._create(
            name=name,
            device=device,
            approvers=approvers,
            contacts=contacts,
            description=description,
            attributes=app_attrs,
            get_if_already_exists=get_if_already_exists
        )
# endregion Applications


# region Application Groups
class _ApplicationGroupBase(FeatureBase):
    def __init__(self, api, class_name):
        super().__init__(api=api)
        self.class_name = class_name
        self.certificate_suffix = f'{class_name.split(maxsplit=1)[0]} App Group'

    def delete(self, application_group: 'Union[Config.Object, str]', dissociate: bool = True):
        """
        Deletes an application group object.

        Args:
            application_group: :ref:`config_object` or :ref:`dn` of the application object.
            dissociate: If ``True``, dissociate all applications in the group from the certificate.
        """
        application_group_dn = self._get_dn(application_group)
        if dissociate:
            consumer_dns, certificate_dn = self._get_applications_in_group(application_group=application_group)
            result = self._api.websdk.Certificates.Dissociate.post(
                certificate_dn=certificate_dn,
                application_dn=consumer_dns,
                delete_orphans=False
            )
            if not result.success:
                raise FeatureException(
                    f'Unable to dissociate the applications in the application group "{application_group.dn}" '
                    f'from {certificate_dn}'
                )
        self._secret_store_delete(object_dn=application_group_dn)
        self._config_delete(object_dn=application_group_dn)

    def get(self, application_group_dn: str, raise_error_if_not_exists: bool = True):
        """
        Args:
            application_group_dn: :ref:`dn` of the application group object.
            raise_error_if_not_exists: Raise an exception if the application :ref:`dn` does not exist.

        Returns:
            :ref:`config_object` of the application.
        """
        return self._get_config_object(
            object_dn=application_group_dn,
            raise_error_if_not_exists=raise_error_if_not_exists
        )

    def get_applications_in_group(self, application_group: 'Union[Config.Object, str]'):
        """
        Args:
            application_group: :ref:`config_object` or :ref:`dn` of the application group.

        Returns:
            List of :ref:`dn`.
        """
        consumer_dns, certificate_dn = self._get_applications_in_group(application_group=application_group)
        return consumer_dns

    def _create(self, application_dns: List[str], certificate: 'Config.Object', attributes: dict = None):
        # Create the Application Group.
        app_group = self._config_create(
            name=f'{certificate.name} - {self.certificate_suffix}',
            parent_folder_dn=certificate.parent,
            config_class=self.class_name,
            attributes=attributes
        )
        # Associate the Application Group :ref:`dn`.
        result = self._api.websdk.Config.WriteDn.post(
            object_dn=certificate.dn,
            attribute_name=X509CertificateAttributes.application_group_dn,
            values=[app_group.dn]
        ).result
        if result.code != 1:
            raise InvalidResultCode(code=result.code, code_description=result.config_result)

        # Associate each individual application.
        result = self._api.websdk.Certificates.Associate.post(
            application_dn=application_dns,
            certificate_dn=certificate.dn,
            push_to_new=False
        )
        if not result.success:
            raise FeatureException(f'Unable to associate the given applications to the certificate "{certificate.dn}".')

        return app_group

    def _get_applications_in_group(self, application_group: 'Union[Config.Object, str]'):
        application_group_dn = self._get_dn(application_group)
        certificate_dn = self._api.websdk.Config.Read.post(
            object_dn=application_group_dn,
            attribute_name=ApplicationGroupAttributes.certificate
        ).values[0]
        consumer_dns = self._api.websdk.Config.Read.post(
            object_dn=certificate_dn,
            attribute_name=X509CertificateAttributes.consumers
        ).values
        return consumer_dns, certificate_dn


@feature('Apache Group')
class ApacheApplicationGroup(_ApplicationGroupBase):
    def __init__(self, api):
        super().__init__(api=api, class_name=Classes.apache_application_group)

    def create(self, applications: 'List[Union[Config.Object, str]]', certificate: 'Union[Config.Object, str]',
               common_data_location: str = None, attributes: dict = None):
        """
        Args:
            applications: List of :ref:`config_object` or :ref:`dn` that will belong to the application group.
            certificate: :ref:`config_object` or :ref:`dn` of the associated certificate.
            common_data_location: The common data location for the application group.
            attributes: Dictionary of attributes to apply to the group explicitly. If not given, then the
                        attributes belonging to the first application in the given list will be assumed to
                        be the shared attributes for the application group.

        Returns:
            :ref:`config_object` of the application.
        """
        application_dns = [self._get_dn(application) for application in applications]
        certificate = self._get_config_object(object_dn=certificate)
        default_attrs = self._api.websdk.Config.ReadAll.post(object_dn=application_dns[0]).name_values
        if not common_data_location:
            try:
                hsm = next(attr.values[0] for attr in default_attrs if
                           attr.name == ApacheApplicationGroupAttributes.private_key_location)
            except StopIteration:
                hsm = None
            if hsm == ApplicationAttributeValues.Apache.PrivateKeyLocation.thales_nshield_hsm:
                common_data_location = '/opt/nfast/bin'
            elif hsm == ApplicationAttributeValues.Apache.PrivateKeyLocation.gemalto_safe_net_hsm:
                common_data_location = '/usr/safenet/lunaclient/bin'

        group_attributes = {
            ApacheApplicationGroupAttributes.client_tools_path            : None,
            ApacheApplicationGroupAttributes.partition_password_credential: None,
            ApacheApplicationGroupAttributes.protection_type              : None,
            ApacheApplicationGroupAttributes.softcard_identifier          : None,
            ApacheApplicationGroupAttributes.private_key_label            : None,
            ApplicationGroupAttributes.common_data_location               : common_data_location,
            ApplicationGroupAttributes.certificate                        : certificate.dn,
            ApplicationGroupAttributes.enrollment_application_dn          : application_dns[0],
        }
        for attribute in default_attrs:
            if attribute.name in group_attributes.keys():
                group_attributes[attribute.name] = attribute.values

        if attributes:
            group_attributes.update(attributes)

        return self._create(application_dns=application_dns, certificate=certificate, attributes=group_attributes)


@feature('PKCS11 Group')
class PKCS11ApplicationGroup(_ApplicationGroupBase):
    """
    .. note::
        The PKCS #11 VSE must be installed in order to use this driver.
    """
    def __init__(self, api):
        super().__init__(api=api, class_name=Classes.pkcs11_application_group)

    def create(self, applications: 'List[Union[Config.Object, str]]', certificate: 'Union[Config.Object, str]',
               attributes: dict = None):
        """
        Args:
            applications: List of :ref:`config_object` or :ref:`dn` that will belong to the application group.
            certificate: :ref:`config_object` or :ref:`dn` of the associated certificate.
            attributes: Dictionary of attributes to apply to the group explicitly. If not given, then the
                        attributes belonging to the first application in the given list will be assumed to
                        be the shared attributes for the application group.

        Returns:
            :ref:`config_object` of the application.
        """
        application_dns = [self._get_dn(application) for application in applications]
        certificate = self._get_config_object(object_dn=certificate)
        default_attrs = self._api.websdk.Config.ReadAll.post(object_dn=application_dns[0]).name_values
        group_attributes = {
            PKCS11ApplicationGroupAttributes.hsm_cka_label_format   : None,
            PKCS11ApplicationGroupAttributes.hsm_requested_cka_label: None,
            PKCS11ApplicationGroupAttributes.hsm_embed_sans_in_csr  : None,
            PKCS11ApplicationGroupAttributes.hsm_import_certificate : None,
            PKCS11ApplicationGroupAttributes.hsm_protection_type    : None,
            PKCS11ApplicationGroupAttributes.hsm_requested_usecase  : None,
            PKCS11ApplicationGroupAttributes.hsm_reverse_subject_dn : None,
            PKCS11ApplicationGroupAttributes.hsm_token_label        : None,
            PKCS11ApplicationGroupAttributes.hsm_token_password     : None,
            ApplicationGroupAttributes.certificate                  : certificate.dn,
            ApplicationGroupAttributes.enrollment_application_dn    : application_dns[0],
        }
        for attribute in default_attrs:
            if attribute.name in group_attributes.keys():
                group_attributes[attribute.name] = attribute.values
        if attributes:
            group_attributes.update(attributes)

        return self._create(application_dns=application_dns, certificate=certificate, attributes=group_attributes)
# endregion Application Groups



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class ACMEAuthorizationAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "ACME Authorization"
	acme_auth_id_type = Attribute('ACME Auth ID Type', min_version='17.2')
	acme_auth_id_value = Attribute('ACME Auth ID Value', min_version='17.2')
	acme_challenge_dn = Attribute('ACME Challenge DN', min_version='21.2')
	acme_challenge_token = Attribute('ACME Challenge Token', min_version='17.2')
	acme_challenge_validated_on = Attribute('ACME Challenge Validated On', min_version='17.2')
	acme_expires = Attribute('ACME Expires', min_version='17.2')
	status = Attribute('Status', min_version='17.2')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.schedule_base import ScheduleBaseAttributes
from pytpp.attributes.top import TopAttributes


class OnboardDiscoveryAttributes(ScheduleBaseAttributes, TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Onboard Discovery"
	amazon_account_ids = Attribute('Amazon Account Ids', min_version='19.2')
	application_type = Attribute('Application Type', min_version='15.2')
	azure_application_id = Attribute('Azure Application Id', min_version='19.2')
	azure_tenant_id = Attribute('Azure Tenant Id', min_version='19.2')
	certificates_placement_folder = Attribute('Certificates Placement Folder', min_version='15.2')
	credential = Attribute('Credential', min_version='18.4')
	default_container = Attribute('Default Container', min_version='15.2')
	device = Attribute('Device', min_version='15.2')
	devices_folder = Attribute('Devices Folder', min_version='15.2')
	disable_installations_not_in_use = Attribute('Disable Installations Not In Use', min_version='21.4')
	driver_name = Attribute('Driver Name', min_version='15.2')
	environment = Attribute('Environment', min_version='20.2')
	extract_private_key = Attribute('Extract Private Key', min_version='15.2')
	in_progress = Attribute('In Progress', min_version='15.2')
	last_run = Attribute('Last Run', min_version='18.2')
	last_update = Attribute('Last Update', min_version='15.2')
	log_debug = Attribute('Log Debug', min_version='19.3')
	onboard_discovery_import_work_to_do = Attribute('Onboard Discovery Import Work To Do', min_version='15.2')
	port = Attribute('Port', min_version='18.1')
	profiles_to_import = Attribute('Profiles To Import', min_version='16.1')
	result_count = Attribute('Result Count', min_version='15.2')
	script_execution_timeout = Attribute('Script Execution Timeout', min_version='20.2')
	script_hash_mismatch_error = Attribute('Script Hash Mismatch Error', min_version='18.4')
	status = Attribute('Status', min_version='15.2')
	stop_requested = Attribute('Stop Requested', min_version='15.2')
	timeout = Attribute('Timeout', min_version='19.2')
	xml_port = Attribute('XML Port', min_version='18.2')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class CodeSigningEnvironmentBaseAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Code Signing Environment Base"
	friendly_name = Attribute('Friendly Name', min_version='20.1')
	ip_address_restriction = Attribute('IP Address Restriction', min_version='20.1')
	key_time_constraint = Attribute('Key Time Constraint', min_version='21.2')
	key_use_authentication = Attribute('Key Use Authentication', min_version='20.1')
	key_use_flow_dn = Attribute('Key Use Flow DN', min_version='20.1')
	per_user = Attribute('Per User', min_version='20.3')
	require_justification = Attribute('Require Justification', min_version='20.1')
	status = Attribute('Status', min_version='20.3')
	template_dn = Attribute('Template DN', min_version='20.1')
	user_key_import = Attribute('User Key Import', min_version='20.3')
